# -*- coding: utf-8 -*-
#
# Copyright © 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluža <jkaluza@redhat.com>
#            Igor Gnatenko <ignatenko@redhat.com>

import collections
import pytest

from module_build_service.mmd_resolver import MMDResolver
from module_build_service import Modulemd


class TestMMDResolver:
    def setup_method(self, test_method):
        self.mmd_resolver = MMDResolver()

    def teardown_method(self, test_method):
        pass

    @staticmethod
    def _make_mmd(nsvc, requires, xmd_buildrequires=None, virtual_streams=None):
        name, stream, version = nsvc.split(":", 2)
        mmd = Modulemd.ModuleStreamV2.new(name, stream)
        mmd.set_summary("foo")
        mmd.set_description("foo")
        for license_ in mmd.get_module_licenses():
            mmd.remove_module_license(license_)
        mmd.add_module_license("GPL")

        xmd = mmd.get_xmd()
        xmd["mbs"] = {}
        xmd["mbs"]["buildrequires"] = {}
        if xmd_buildrequires:
            for ns in xmd_buildrequires:
                n, s = ns.split(":")
                xmd["mbs"]["buildrequires"][n] = {"stream": s}
        if virtual_streams:
            xmd["mbs"]["virtual_streams"] = virtual_streams
        mmd.set_xmd(xmd)

        if ":" in version:
            version, context = version.split(":")
            mmd.set_context(context)
            add_requires = Modulemd.Dependencies.add_runtime_stream
            add_empty_requires = Modulemd.Dependencies.set_empty_runtime_dependencies_for_module
        else:
            add_requires = Modulemd.Dependencies.add_buildtime_stream
            add_empty_requires = Modulemd.Dependencies.set_empty_buildtime_dependencies_for_module
        mmd.set_version(int(version))

        if not isinstance(requires, list):
            requires = [requires]
        else:
            requires = requires

        for reqs in requires:
            deps = Modulemd.Dependencies()
            for req_name, req_streams in reqs.items():
                if req_streams == []:
                    add_empty_requires(deps, req_name)
                else:
                    for req_stream in req_streams:
                        add_requires(deps, req_name, req_stream)
            mmd.add_dependencies(deps)

        return mmd

    @pytest.mark.parametrize(
        "deps, expected",
        (
            ([], "None"),
            ([{"x": []}], "module(x)"),
            ([{"x": ["1"]}], "(module(x) with module(x:1))"),
            ([{"x": ["-1"]}], "(module(x) without module(x:1))"),
            ([{"x": ["1", "2"]}], "(module(x) with (module(x:1) or module(x:2)))"),
            ([{"x": ["-1", "2"]}], "(module(x) with module(x:2))"),
            ([{"x": [], "y": []}], "(module(x) and module(y))"),
            ([{"x": []}, {"y": []}], "(module(x) or module(y))"),
        ),
    )
    def test_deps2reqs(self, deps, expected):
        # Sort by keys here to avoid unordered dicts
        deps = [collections.OrderedDict(sorted(dep.items())) for dep in deps]
        reqs = self.mmd_resolver._deps2reqs(deps)
        assert str(reqs) == expected

    @pytest.mark.parametrize(
        "buildrequires, expected",
        (
            ({"platform": []}, [[["platform:f28:0:c0:x86_64"], ["platform:f29:0:c0:x86_64"]]]),
            ({"platform": ["f28"]}, [[["platform:f28:0:c0:x86_64"]]]),
            ({"platform": ["-f28"]}, [[["platform:f29:0:c0:x86_64"]]]),
            (
                {"gtk": [], "qt": []},
                [
                    [
                        ["gtk:3:0:c8:x86_64", "qt:4:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                        ["gtk:4:0:c8:x86_64", "qt:4:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                        ["gtk:3:0:c8:x86_64", "qt:5:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                        ["gtk:4:0:c8:x86_64", "qt:5:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                    ]
                ],
            ),
            (
                {"gtk": [], "qt": [], "platform": []},
                [
                    [
                        ["gtk:3:0:c8:x86_64", "qt:4:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                        ["gtk:4:0:c8:x86_64", "qt:4:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                        ["gtk:3:0:c8:x86_64", "qt:5:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                        ["gtk:4:0:c8:x86_64", "qt:5:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                        ["gtk:3:0:c9:x86_64", "qt:4:0:c9:x86_64", "platform:f29:0:c0:x86_64"],
                        ["gtk:4:0:c9:x86_64", "qt:4:0:c9:x86_64", "platform:f29:0:c0:x86_64"],
                        ["gtk:3:0:c9:x86_64", "qt:5:0:c9:x86_64", "platform:f29:0:c0:x86_64"],
                        ["gtk:4:0:c9:x86_64", "qt:5:0:c9:x86_64", "platform:f29:0:c0:x86_64"],
                    ]
                ],
            ),
            (
                [{"qt": [], "platform": ["f28"]}, {"gtk": [], "platform": ["-f28"]}],
                [
                    [
                        ["qt:4:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                        ["qt:5:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                    ],
                    [
                        ["gtk:3:0:c9:x86_64", "platform:f29:0:c0:x86_64"],
                        ["gtk:4:0:c9:x86_64", "platform:f29:0:c0:x86_64"],
                    ],
                ],
            ),
            (
                {"mess": []},
                [[["mess:1:0:c0:x86_64", "gtk:3:0:c8:x86_64", "platform:f28:0:c0:x86_64"]]],
            ),
            (
                {"mess": [], "platform": []},
                [
                    [
                        ["mess:1:0:c0:x86_64", "gtk:3:0:c8:x86_64", "platform:f28:0:c0:x86_64"],
                        ["mess:1:0:c0:x86_64", "gtk:4:0:c9:x86_64", "platform:f29:0:c0:x86_64"],
                    ]
                ],
            ),
        ),
    )
    def test_solve(self, buildrequires, expected):
        modules = (
            ("platform:f28:0:c0", {}),
            ("platform:f29:0:c0", {}),
            ("gtk:3:0:c8", {"platform": ["f28"]}),
            ("gtk:3:0:c9", {"platform": ["f29"]}),
            ("gtk:4:0:c8", {"platform": ["f28"]}),
            ("gtk:4:0:c9", {"platform": ["f29"]}),
            ("qt:4:0:c8", {"platform": ["f28"]}),
            ("qt:4:0:c9", {"platform": ["f29"]}),
            ("qt:5:0:c8", {"platform": ["f28"]}),
            ("qt:5:0:c9", {"platform": ["f29"]}),
            (
                "mess:1:0:c0",
                [{"gtk": ["3"], "platform": ["f28"]}, {"gtk": ["4"], "platform": ["-f28"]}],
            ),
        )
        for n, req in modules:
            self.mmd_resolver.add_modules(self._make_mmd(n, req))

        app = self._make_mmd("app:1:0", buildrequires)
        expanded = self.mmd_resolver.solve(app)

        expected = set(
            frozenset(["app:1:0:%d:src" % c] + e) for c, exp in enumerate(expected) for e in exp
        )

        assert expanded == expected

    @pytest.mark.parametrize(
        "buildrequires, xmd_buildrequires, expected",
        (
            # BR all platform streams -> build for all platform streams.
            (
                {"platform": []},
                {},
                [
                    [
                        ["platform:el8.2.0.z:0:c0:x86_64"],
                        ["platform:el8.1.0:0:c0:x86_64"],
                        ["platform:el8.0.0:0:c0:x86_64"],
                        ["platform:el7.6.0:0:c0:x86_64"],
                    ]
                ],
            ),
            # BR "el8" platform stream -> build for all el8 platform streams.
            (
                {"platform": ["el8"]},
                {},
                [
                    [
                        ["platform:el8.2.0.z:0:c0:x86_64"],
                        ["platform:el8.1.0:0:c0:x86_64"],
                        ["platform:el8.0.0:0:c0:x86_64"],
                    ]
                ],
            ),
            # BR "el8.1.0" platfrom stream -> build just for el8.1.0.
            ({"platform": ["el8"]}, ["platform:el8.1.0"], [[["platform:el8.1.0:0:c0:x86_64"]]]),
            # BR platform:el8.1.0 and gtk:3, which is not built against el8.1.0,
            # but it is built only against el8.0.0 -> cherry-pick gtk:3 from el8.0.0
            # and build once against platform:el8.1.0.
            (
                {"platform": ["el8"], "gtk": ["3"]},
                ["platform:el8.1.0"],
                [[["platform:el8.1.0:0:c0:x86_64", "gtk:3:0:c8:x86_64"]]],
            ),
            # BR platform:el8.2.0 and gtk:3, this time gtk:3 build against el8.2.0 exists
            # -> use both platform and gtk from el8.2.0 and build once.
            (
                {"platform": ["el8"], "gtk": ["3"]},
                ["platform:el8.2.0.z"],
                [[["platform:el8.2.0.z:0:c0:x86_64", "gtk:3:1:c8:x86_64"]]],
            ),
            # BR platform:el8.2.0 and mess:1 which is built against platform:el8.1.0 and
            # requires gtk:3 which is built against platform:el8.2.0 and platform:el8.0.0
            # -> Use platform:el8.2.0 and
            # -> cherry-pick mess:1 from el8.1.0 and
            # -> use gtk:3:1 from el8.2.0.
            (
                {"platform": ["el8"], "mess": ["1"]},
                ["platform:el8.2.0.z"],
                [[["platform:el8.2.0.z:0:c0:x86_64", "mess:1:0:c0:x86_64", "gtk:3:1:c8:x86_64"]]],
            ),
            # BR platform:el8.1.0 and mess:1 which is built against platform:el8.1.0 and
            # requires gtk:3 which is built against platform:el8.2.0 and platform:el8.0.0
            # -> Use platform:el8.1.0 and
            # -> Used mess:1 from el8.1.0 and
            # -> cherry-pick gtk:3:0 from el8.0.0.
            (
                {"platform": ["el8"], "mess": ["1"]},
                ["platform:el8.1.0"],
                [[["platform:el8.1.0:0:c0:x86_64", "mess:1:0:c0:x86_64", "gtk:3:0:c8:x86_64"]]],
            ),
            # BR platform:el8.0.0 and mess:1 which is built against platform:el8.1.0 and
            # requires gtk:3 which is built against platform:el8.2.0 and platform:el8.0.0
            # -> No valid combination, because mess:1 is only available in el8.1.0 and later.
            ({"platform": ["el8"], "mess": ["1"]}, ["platform:el8.0.0"], []),
            # This is undefined... it might build just once against latest platform or
            # against all the platforms... we don't know
            # ({"platform": ["el8"], "gtk": ["3"]}, {}, [
            #     [["platform:el8.2.0:0:c0:x86_64", "gtk:3:1:c8:x86_64"]],
            # ]),
        ),
    )
    def test_solve_virtual_streams(self, buildrequires, xmd_buildrequires, expected):
        modules = (
            # (nsvc, buildrequires, expanded_buildrequires, virtual_streams)
            ("platform:el8.0.0:0:c0", {}, {}, ["el8"]),
            ("platform:el8.1.0:0:c0", {}, {}, ["el8"]),
            ("platform:el8.2.0.z:0:c0", {}, {}, ["el8"]),
            ("platform:el7.6.0:0:c0", {}, {}, ["el7"]),
            ("gtk:3:0:c8", {"platform": ["el8"]}, {"platform:el8.0.0"}, None),
            ("gtk:3:1:c8", {"platform": ["el8"]}, {"platform:el8.2.0.z"}, None),
            ("mess:1:0:c0", [{"gtk": ["3"], "platform": ["el8"]}], {"platform:el8.1.0"}, None),
        )
        for n, req, xmd_br, virtual_streams in modules:
            self.mmd_resolver.add_modules(self._make_mmd(n, req, xmd_br, virtual_streams))

        app = self._make_mmd("app:1:0", buildrequires, xmd_buildrequires)
        if not expected:
            with pytest.raises(RuntimeError):
                self.mmd_resolver.solve(app)
            return
        else:
            expanded = self.mmd_resolver.solve(app)

        expected = set(
            frozenset(["app:1:0:%d:src" % c] + e) for c, exp in enumerate(expected) for e in exp)

        assert expanded == expected

    @pytest.mark.parametrize(
        "app_buildrequires, modules, err_msg_regex",
        (
            # app --br--> gtk:1 --req--> bar:1* ---req---> platform:f29
            #    \--br--> foo:1 --req--> bar:2* ---req--/
            (
                {"gtk": "1", "foo": "1"},
                (
                    ("platform:f29:0:c0", {}),
                    ("gtk:1:1:c01", {"bar": ["1"]}),
                    ("bar:1:0:c02", {"platform": ["f29"]}),
                    ("foo:1:1:c03", {"bar": ["2"]}),
                    ("bar:2:0:c04", {"platform": ["f29"]}),
                ),
                "bar:1:0:c02 and bar:2:0:c04",
            ),
            # app --br--> gtk:1 --req--> bar:1* ----------req----------> platform:f29
            #    \--br--> foo:1 --req--> baz:1 --req--> bar:2* --req--/
            (
                {"gtk": "1", "foo": "1"},
                (
                    ("platform:f29:0:c0", {}),
                    ("gtk:1:1:c01", {"bar": ["1"]}),
                    ("bar:1:0:c02", {"platform": ["f29"]}),
                    ("foo:1:1:c03", {"baz": ["1"]}),
                    ("baz:1:1:c04", {"bar": ["2"]}),
                    ("bar:2:0:c05", {"platform": ["f29"]}),
                ),
                "bar:1:0:c02 and bar:2:0:c05",
            ),
            # Test multiple conflicts pairs are detected.
            # app --br--> gtk:1 --req--> bar:1* ---------req-----------\
            #    \--br--> foo:1 --req--> baz:1 --req--> bar:2* ---req---> platform:f29
            #    \--br--> pkga:1 --req--> perl:5' -------req-----------/
            #    \--br--> pkgb:1 --req--> perl:6' -------req-----------/
            (
                {"gtk": "1", "foo": "1", "pkga": "1", "pkgb": "1"},
                (
                    ("platform:f29:0:c0", {}),
                    ("gtk:1:1:c01", {"bar": ["1"]}),
                    ("bar:1:0:c02", {"platform": ["f29"]}),
                    ("foo:1:1:c03", {"baz": ["1"]}),
                    ("baz:1:1:c04", {"bar": ["2"]}),
                    ("bar:2:0:c05", {"platform": ["f29"]}),
                    ("pkga:1:0:c06", {"perl": ["5"]}),
                    ("perl:5:0:c07", {"platform": ["f29"]}),
                    ("pkgb:1:0:c08", {"perl": ["6"]}),
                    ("perl:6:0:c09", {"platform": ["f29"]}),
                ),
                # MMD Resolver should still catch a conflict
                "bar:1:0:c02 and bar:2:0:c05",
            ),
        ),
    )
    def test_solve_stream_conflicts(self, app_buildrequires, modules, err_msg_regex):
        for n, req in modules:
            self.mmd_resolver.add_modules(self._make_mmd(n, req))

        app = self._make_mmd("app:1:0", app_buildrequires)

        with pytest.raises(RuntimeError, match=err_msg_regex):
            self.mmd_resolver.solve(app)

    def test_solve_new_platform(self,):
        """
        Tests that MMDResolver works in case we add new platform and there is
        modular dependency of input module which is built only against old
        platforms and in the same time the input module wants to be built
        with all available platforms.
        """
        modules = (
            ("platform:f28:0:c0", {}),
            ("platform:f29:0:c0", {}),
            ("platform:f30:0:c0", {}),
            ("gtk:3:0:c8", {"platform": ["f28"]}),
            ("gtk:3:0:c9", {"platform": ["f29"]}),
        )
        for n, req in modules:
            self.mmd_resolver.add_modules(self._make_mmd(n, req))

        app = self._make_mmd("app:1:0", {"platform": [], "gtk": ["3"]})
        expanded = self.mmd_resolver.solve(app)

        # Build only against f28 and f29, because "gtk:3" is not built against f30.
        expected = set([
            frozenset(["gtk:3:0:c8:x86_64", "app:1:0:0:src", "platform:f28:0:c0:x86_64"]),
            frozenset(["gtk:3:0:c9:x86_64", "app:1:0:0:src", "platform:f29:0:c0:x86_64"]),
        ])

        assert expanded == expected

    def test_solve_requires_any_platform(self,):
        """
        Tests that MMDResolver finds out the buildrequired module `foo` even if
        it was built on newer platform stream, but can run on any platform stream.
        """
        modules = (
            ("platform:f28:0:c0", {}, {}),
            ("platform:f29:0:c0", {}, {}),
            ("platform:f30:0:c0", {}, {}),
            ("foo:1:0:c8", {"platform": []}, ["platform:f29"]),
        )
        for n, req, xmd_req in modules:
            self.mmd_resolver.add_modules(self._make_mmd(n, req, xmd_req))

        app = self._make_mmd("app:1:0", {"platform": ["f28"], "foo": ["1"]})
        expanded = self.mmd_resolver.solve(app)

        expected = set([
            frozenset(["foo:1:0:c8:x86_64", "app:1:0:0:src", "platform:f28:0:c0:x86_64"]),
        ])

        assert expanded == expected



# Copyright (c) 2019  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

import os
from datetime import datetime

import module_build_service.resolver as mbs_resolver
from module_build_service import db
from module_build_service.utils.general import import_mmd, load_mmd_file, mmd_to_str
from module_build_service.models import ModuleBuild
import tests


base_dir = os.path.join(os.path.dirname(__file__), "..")


class TestLocalResolverModule:
    def setup_method(self):
        tests.reuse_component_init_data()

    def test_get_buildrequired_modulemds(self):
        mmd = load_mmd_file(os.path.join(base_dir, "staged_data", "platform.yaml"))
        mmd = mmd.copy(mmd.get_module_name(), "f8")
        import_mmd(db.session, mmd)
        platform_f8 = ModuleBuild.query.filter_by(stream="f8").one()
        mmd = mmd.copy("testmodule", "master")
        mmd.set_version(20170109091357)
        mmd.set_context("123")
        build = ModuleBuild(
            name="testmodule",
            stream="master",
            version=20170109091357,
            state=5,
            build_context="dd4de1c346dcf09ce77d38cd4e75094ec1c08ec3",
            runtime_context="ec4de1c346dcf09ce77d38cd4e75094ec1c08ef7",
            context="7c29193d",
            koji_tag="module-testmodule-master-20170109091357-7c29193d",
            scmurl="https://src.stg.fedoraproject.org/modules/testmodule.git?#ff1ea79",
            batch=3,
            owner="Dr. Pepper",
            time_submitted=datetime(2018, 11, 15, 16, 8, 18),
            time_modified=datetime(2018, 11, 15, 16, 19, 35),
            rebuild_strategy="changed-and-after",
            modulemd=mmd_to_str(mmd),
        )
        db.session.add(build)
        db.session.commit()

        resolver = mbs_resolver.GenericResolver.create(tests.conf, backend="local")
        result = resolver.get_buildrequired_modulemds(
            "testmodule", "master", platform_f8.mmd().get_nsvc())
        nsvcs = set([m.get_nsvc() for m in result])
        assert nsvcs == set(
            ["testmodule:master:20170109091357:9c690d0e", "testmodule:master:20170109091357:123"])



# -*- coding: utf-8 -*-
# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>
# Written by Matt Prahl <mprahl@redhat.com>

"""Auth system based on the client certificate and FAS account"""
import json
import os
from socket import gethostname
import ssl

import requests
import kerberos
from flask import Response, g

# Starting with Flask 0.9, the _app_ctx_stack is the correct one,
# before that we need to use the _request_ctx_stack.
try:
    from flask import _app_ctx_stack as stack
except ImportError:
    from flask import _request_ctx_stack as stack
from werkzeug.exceptions import Unauthorized as FlaskUnauthorized
from dogpile.cache import make_region

from module_build_service.errors import Unauthorized, Forbidden
from module_build_service import app, log, conf

try:
    import ldap3
except ImportError:
    log.warning("ldap3 import not found.  ldap/krb disabled.")


client_secrets = None
region = make_region().configure("dogpile.cache.memory")


def _json_loads(content):
    if not isinstance(content, str):
        content = content.decode("utf-8")
    return json.loads(content)


def _load_secrets():
    global client_secrets
    if client_secrets:
        return

    if "OIDC_CLIENT_SECRETS" not in app.config:
        raise Forbidden("OIDC_CLIENT_SECRETS must be set in server config.")

    secrets = _json_loads(open(app.config["OIDC_CLIENT_SECRETS"], "r").read())
    client_secrets = list(secrets.values())[0]


def _get_token_info(token):
    """
    Asks the token_introspection_uri for the validity of a token.
    """
    if not client_secrets:
        return None

    request = {
        "token": token,
        "token_type_hint": "Bearer",
        "client_id": client_secrets["client_id"],
        "client_secret": client_secrets["client_secret"],
    }
    headers = {"Content-type": "application/x-www-form-urlencoded"}

    resp = requests.post(client_secrets["token_introspection_uri"], data=request, headers=headers)
    return resp.json()


def _get_user_info(token):
    """
    Asks the userinfo_uri for more information on a user.
    """
    if not client_secrets:
        return None

    headers = {"authorization": "Bearer " + token}
    resp = requests.get(client_secrets["userinfo_uri"], headers=headers)
    return resp.json()


def get_user_oidc(request):
    """
    Returns the client's username and groups based on the OIDC token provided.
    """
    _load_secrets()

    if "authorization" not in request.headers:
        raise Unauthorized("No 'authorization' header found.")

    header = request.headers["authorization"].strip()
    prefix = "Bearer "
    if not header.startswith(prefix):
        raise Unauthorized("Authorization headers must start with %r" % prefix)

    token = header[len(prefix):].strip()
    try:
        data = _get_token_info(token)
    except Exception as e:
        error = "Cannot verify OIDC token: %s" % str(e)
        log.exception(error)
        raise Exception(error)

    if not data or "active" not in data or not data["active"]:
        raise Unauthorized("OIDC token invalid or expired.")

    if "OIDC_REQUIRED_SCOPE" not in app.config:
        raise Forbidden("OIDC_REQUIRED_SCOPE must be set in server config.")

    presented_scopes = data["scope"].split(" ")
    required_scopes = [
        "openid",
        "https://id.fedoraproject.org/scope/groups",
        app.config["OIDC_REQUIRED_SCOPE"],
    ]
    for scope in required_scopes:
        if scope not in presented_scopes:
            raise Unauthorized("Required OIDC scope %r not present: %r" % (scope, presented_scopes))

    try:
        extended_data = _get_user_info(token)
    except Exception as e:
        error = "Cannot verify determine user groups:  %s" % str(e)
        log.exception(error)
        raise Exception(error)

    username = data["username"]
    # If the user is part of the whitelist, then the group membership check is skipped
    if username in conf.allowed_users:
        groups = set()
    else:
        try:
            groups = set(extended_data["groups"])
        except Exception as e:
            error = "Could not find groups in UserInfo from OIDC %s" % str(e)
            log.exception(extended_data)
            raise Exception(error)

    return username, groups


# Insired by https://pagure.io/waiverdb/blob/master/f/waiverdb/auth.py which was
# inspired by https://github.com/mkomitee/flask-kerberos/blob/master/flask_kerberos.py
class KerberosAuthenticate(object):
    def __init__(self):
        if conf.kerberos_http_host:
            hostname = conf.kerberos_http_host
        else:
            hostname = gethostname()
        self.service_name = "HTTP@{0}".format(hostname)

        # If the config specifies a keytab to use, then override the KRB5_KTNAME
        # environment variable
        if conf.kerberos_keytab:
            os.environ["KRB5_KTNAME"] = conf.kerberos_keytab

        if "KRB5_KTNAME" in os.environ:
            try:
                principal = kerberos.getServerPrincipalDetails("HTTP", hostname)
            except kerberos.KrbError as error:
                raise Unauthorized('Kerberos: authentication failed with "{0}"'.format(str(error)))

            log.debug('Kerberos: server is identifying as "{0}"'.format(principal))
        else:
            raise Unauthorized(
                'Kerberos: set the config value of "KERBEROS_KEYTAB" or the '
                'environment variable "KRB5_KTNAME" to your keytab file'
            )

    def _gssapi_authenticate(self, token):
        """
        Performs GSSAPI Negotiate Authentication
        On success also stashes the server response token for mutual authentication
        at the top of request context with the name kerberos_token, along with the
        authenticated user principal with the name kerberos_user.
        """
        state = None
        ctx = stack.top
        try:
            rc, state = kerberos.authGSSServerInit(self.service_name)
            if rc != kerberos.AUTH_GSS_COMPLETE:
                log.error("Kerberos: unable to initialize server context")
                return None

            rc = kerberos.authGSSServerStep(state, token)
            if rc == kerberos.AUTH_GSS_COMPLETE:
                log.debug("Kerberos: completed GSSAPI negotiation")
                ctx.kerberos_token = kerberos.authGSSServerResponse(state)
                ctx.kerberos_user = kerberos.authGSSServerUserName(state)
                return rc
            elif rc == kerberos.AUTH_GSS_CONTINUE:
                log.debug("Kerberos: continuing GSSAPI negotiation")
                return kerberos.AUTH_GSS_CONTINUE
            else:
                log.debug("Kerberos: unable to step server context")
                return None
        except kerberos.GSSError as error:
            log.error("Kerberos: unable to authenticate: {0}".format(str(error)))
            return None
        finally:
            if state:
                kerberos.authGSSServerClean(state)

    def process_request(self, token):
        """
        Authenticates the current request using Kerberos.
        """
        kerberos_user = None
        kerberos_token = None
        ctx = stack.top
        rc = self._gssapi_authenticate(token)
        if rc == kerberos.AUTH_GSS_COMPLETE:
            kerberos_user = ctx.kerberos_user
            kerberos_token = ctx.kerberos_token
        elif rc != kerberos.AUTH_GSS_CONTINUE:
            raise Forbidden("Invalid Kerberos ticket")

        return kerberos_user, kerberos_token


def get_user_kerberos(request):
    user = None
    if "Authorization" not in request.headers:
        response = Response("Unauthorized", 401, {"WWW-Authenticate": "Negotiate"})
        exc = FlaskUnauthorized()
        # For some reason, certain versions of werkzeug raise an exception when passing `response`
        # in the constructor. This is a work-around.
        exc.response = response
        raise exc
    header = request.headers.get("Authorization")
    token = "".join(header.strip().split()[1:])
    user, kerberos_token = KerberosAuthenticate().process_request(token)
    # Remove the realm
    user = user.split("@")[0]
    # If the user is part of the whitelist, then the group membership check is skipped
    if user in conf.allowed_users:
        groups = []
    else:
        groups = get_ldap_group_membership(user)
    return user, set(groups)


@region.cache_on_arguments()
def get_ldap_group_membership(uid):
    """ Small wrapper on getting the group membership so that we can use caching
    :param uid: a string of the uid of the user
    :return: a list of groups the user is a member of
    """
    ldap_con = Ldap()
    return ldap_con.get_user_membership(uid)


class Ldap(object):
    """ A class that handles LDAP connections and queries
    """

    connection = None
    base_dn = None

    def __init__(self):
        if not conf.ldap_uri:
            raise Forbidden("LDAP_URI must be set in server config.")
        if conf.ldap_groups_dn:
            self.base_dn = conf.ldap_groups_dn
        else:
            raise Forbidden("LDAP_GROUPS_DN must be set in server config.")

        if conf.ldap_uri.startswith("ldaps://"):
            tls = ldap3.Tls(
                ca_certs_file="/etc/pki/tls/certs/ca-bundle.crt", validate=ssl.CERT_REQUIRED)
            server = ldap3.Server(conf.ldap_uri, use_ssl=True, tls=tls)
        else:
            server = ldap3.Server(conf.ldap_uri)
        self.connection = ldap3.Connection(server)
        try:
            self.connection.open()
        except ldap3.core.exceptions.LDAPSocketOpenError as error:
            log.error(
                'The connection to "{0}" failed. The following error was raised: {1}'.format(
                    conf.ldap_uri, str(error)))
            raise Forbidden(
                "The connection to the LDAP server failed. Group membership couldn't be obtained.")

    def get_user_membership(self, uid):
        """ Gets the group membership of a user
        :param uid: a string of the uid of the user
        :return: a list of common names of the posixGroups the user is a member of
        """
        ldap_filter = "(memberUid={0})".format(uid)
        # Only get the groups in the base container/OU
        self.connection.search(
            self.base_dn, ldap_filter, search_scope=ldap3.LEVEL, attributes=["cn"])
        groups = self.connection.response
        try:
            return [group["attributes"]["cn"][0] for group in groups]
        except KeyError:
            log.exception(
                "The LDAP groups could not be determined based on the search results "
                'of "{0}"'.format(str(groups)))
            return []


def get_user(request):
    """ Authenticates the user and returns the username and group name
    :param request: a Flask request
    :return: a tuple with a string representing the user name and a set with the user's group
    membership such as ('mprahl', {'factory2', 'devel'})
    """
    if conf.no_auth is True:
        log.debug("Authorization is disabled.")
        return "anonymous", {"packager"}

    if "user" not in g and "groups" not in g:
        get_user_func_name = "get_user_{0}".format(conf.auth_method)
        get_user_func = globals().get(get_user_func_name)
        if not get_user_func:
            raise RuntimeError('The function "{0}" is not implemented'.format(get_user_func_name))
        g.user, g.groups = get_user_func(request)
    return g.user, g.groups



# Copyright (c) 2019  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>

import pytest

from mock import call, patch, Mock
from sqlalchemy import func

from module_build_service import conf, db
from module_build_service.models import BUILD_STATES, ModuleBuild
from module_build_service.scheduler.consumer import MBSConsumer
from module_build_service.scheduler.handlers.greenwave import get_corresponding_module_build
from module_build_service.scheduler.handlers.greenwave import decision_update
from tests import clean_database, make_module


class TestGetCorrespondingModuleBuild:
    """Test get_corresponding_module_build"""

    def setup_method(self, method):
        clean_database()

    @patch("module_build_service.builder.KojiModuleBuilder.KojiClientSession")
    def test_module_build_nvr_does_not_exist_in_koji(self, ClientSession):
        ClientSession.return_value.getBuild.return_value = None

        assert get_corresponding_module_build(db.session, "n-v-r") is None

    @pytest.mark.parametrize(
        "build_info",
        [
            # Build info does not have key extra
            {"id": 1000, "name": "ed"},
            # Build info contains key extra, but it is not for the module build
            {"extra": {"submitter": "osbs", "image": {}}},
            # Key module_build_service_id is missing
            {"extra": {"typeinfo": {"module": {}}}},
        ],
    )
    @patch("module_build_service.builder.KojiModuleBuilder.KojiClientSession")
    def test_cannot_find_module_build_id_from_build_info(self, ClientSession, build_info):
        ClientSession.return_value.getBuild.return_value = build_info

        assert get_corresponding_module_build(db.session, "n-v-r") is None

    @patch("module_build_service.builder.KojiModuleBuilder.KojiClientSession")
    def test_corresponding_module_build_id_does_not_exist_in_db(self, ClientSession):
        fake_module_build_id, = db.session.query(func.max(ModuleBuild.id)).first()

        ClientSession.return_value.getBuild.return_value = {
            "extra": {"typeinfo": {"module": {"module_build_service_id": fake_module_build_id + 1}}}
        }

        assert get_corresponding_module_build(db.session, "n-v-r") is None

    @patch("module_build_service.builder.KojiModuleBuilder.KojiClientSession")
    def test_find_the_module_build(self, ClientSession):
        expected_module_build = (
            db.session.query(ModuleBuild).filter(ModuleBuild.name == "platform").first()
        )

        ClientSession.return_value.getBuild.return_value = {
            "extra": {"typeinfo": {"module": {"module_build_service_id": expected_module_build.id}}}
        }

        build = get_corresponding_module_build(db.session, "n-v-r")

        assert expected_module_build.id == build.id
        assert expected_module_build.name == build.name


class TestDecisionUpdateHandler:
    """Test handler decision_update"""

    @patch("module_build_service.scheduler.handlers.greenwave.log")
    def test_decision_context_is_not_match(self, log):
        msg = Mock(msg_id="msg-id-1", decision_context="bodhi_update_push_testing")
        decision_update(conf, db.session, msg)
        log.debug.assert_called_once_with(
            'Skip Greenwave message %s as MBS only handles messages with the decision context "%s"',
            "msg-id-1",
            "test_dec_context"
        )

    @patch("module_build_service.scheduler.handlers.greenwave.log")
    def test_not_satisfy_policies(self, log):
        msg = Mock(
            msg_id="msg-id-1",
            decision_context="test_dec_context",
            policies_satisfied=False,
            subject_identifier="pkg-0.1-1.c1",
        )
        decision_update(conf, db.session, msg)
        log.debug.assert_called_once_with(
            "Skip to handle module build %s because it has not satisfied Greenwave policies.",
            msg.subject_identifier,
        )

    @patch("module_build_service.messaging.publish")
    @patch("module_build_service.builder.KojiModuleBuilder.KojiClientSession")
    def test_transform_from_done_to_ready(self, ClientSession, publish):
        clean_database()

        # This build should be queried and transformed to ready state
        module_build = make_module("pkg:0.1:1:c1", requires_list={"platform": "el8"})
        module_build.transition(
            conf, BUILD_STATES["done"], "Move to done directly for running test."
        )

        # Assert this call below
        first_publish_call = call(
            service="mbs",
            topic="module.state.change",
            msg=module_build.json(show_tasks=False),
            conf=conf,
        )

        db.session.refresh(module_build)

        ClientSession.return_value.getBuild.return_value = {
            "extra": {"typeinfo": {"module": {"module_build_service_id": module_build.id}}}
        }

        msg = {
            "msg_id": "msg-id-1",
            "topic": "org.fedoraproject.prod.greenwave.decision.update",
            "msg": {
                "decision_context": "test_dec_context",
                "policies_satisfied": True,
                "subject_identifier": "pkg-0.1-1.c1",
            },
        }
        hub = Mock(config={"validate_signatures": False})
        consumer = MBSConsumer(hub)
        consumer.consume(msg)

        # Load module build again to check its state is moved correctly
        module_build = (
            db.session.query(ModuleBuild).filter(ModuleBuild.id == module_build.id).first())

        assert BUILD_STATES["ready"] == module_build.state

        publish.assert_has_calls([
            first_publish_call,
            call(
                service="mbs",
                topic="module.state.change",
                msg=module_build.json(show_tasks=False),
                conf=conf,
            ),
        ])



# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Matt Prahl <mprahl@redhat.com

import os
from datetime import datetime, timedelta
from mock import patch
import time
import hashlib
from traceback import extract_stack
from module_build_service.utils import to_text_type, load_mmd_file

import koji
import module_build_service
from module_build_service import db
from module_build_service.utils import get_rpm_release, import_mmd, mmd_to_str
from module_build_service.config import init_config
from module_build_service.models import (
    ModuleBuild, ComponentBuild, VirtualStream, make_session, BUILD_STATES,
)
from module_build_service import Modulemd


base_dir = os.path.dirname(__file__)
app = module_build_service.app
conf = init_config(app)


def read_staged_data(yaml_name):
    """Read module YAML content from staged_data directory

    :param str yaml_name: name of YAML file without extension ``.yaml``.
    :return: module YAML file's content.
    :rtype: str
    :raises ValueError: if specified module YAML file does not exist in
        staged_data directory.
    """
    filename = os.path.join(base_dir, "staged_data", "{}.yaml".format(yaml_name))
    if not os.path.exists(filename):
        raise ValueError("Staged data {}.yaml does not exist.".format(yaml_name))
    with open(filename, "r") as mmd:
        return to_text_type(mmd.read())


def patch_config():
    # add test builders for all resolvers
    with_test_builders = dict()
    for k, v in module_build_service.config.SUPPORTED_RESOLVERS.items():
        v["builders"].extend(["test", "testlocal"])
        with_test_builders[k] = v
    patch("module_build_service.config.SUPPORTED_RESOLVERS", with_test_builders)


patch_config()


def patch_zeromq_time_sleep():
    """
    We use moksha.hub in some tests. We used dummy zerombq backend which
    connects to /dev/null, but zeromq.py contains time.sleep(1) to ensure
    that sockets are listening properly. This is not needed for our dummy
    use-case and it slows down tests.

    This method patches time.sleep called from "zeromq.py" file to be noop,
    but calls the real time.sleep otherwise.
    """
    global _orig_time_sleep
    _orig_time_sleep = time.sleep

    def mocked_time_sleep(n):
        global _orig_time_sleep
        if n == 1:
            tb = extract_stack()
            try:
                if tb[-4][0].endswith("zeromq.py"):
                    return
            except IndexError:
                pass
        _orig_time_sleep(n)

    ts = patch("time.sleep").start()
    ts.side_effect = mocked_time_sleep


patch_zeromq_time_sleep()


def clean_database(add_platform_module=True, add_default_arches=True):
    db.session.commit()
    db.drop_all()
    db.create_all()

    if add_default_arches:
        arch_obj = module_build_service.models.ModuleArch(name="x86_64")
        db.session.add(arch_obj)
        db.session.commit()

    if add_platform_module:
        mmd = load_mmd_file(os.path.join(base_dir, "staged_data", "platform.yaml"))
        import_mmd(db.session, mmd)


def init_data(data_size=10, contexts=False, multiple_stream_versions=None, scratch=False):
    """
    Creates data_size * 3 modules in database in different states and
    with different component builds. See _populate_data for more info.

    :param bool contexts: If True, multiple streams and contexts in each stream
        are generated for 'nginx' module.
    :param list/bool multiple_stream_versions: If true, multiple base modules with
        difference stream versions are generated. If set to list, the list defines
        the generated base module streams.
    """
    clean_database()
    if multiple_stream_versions:
        if multiple_stream_versions is True:
            multiple_stream_versions = ["f28.0.0", "f29.0.0", "f29.1.0", "f29.2.0"]
        mmd = load_mmd_file(os.path.join(base_dir, "staged_data", "platform.yaml"))
        for stream in multiple_stream_versions:
            mmd = mmd.copy("platform", stream)

            # Set the virtual_streams based on "fXY" to mark the platform streams
            # with the same major stream_version compatible.
            xmd = mmd.get_xmd()
            xmd["mbs"]["virtual_streams"] = [stream[:3]]
            mmd.set_xmd(xmd)
            import_mmd(db.session, mmd)

            # Just to possibly confuse tests by adding another base module.
            mmd = mmd.copy("bootstrap", stream)
            import_mmd(db.session, mmd)
    with make_session(conf) as session:
        _populate_data(session, data_size, contexts=contexts, scratch=scratch)


def _populate_data(session, data_size=10, contexts=False, scratch=False):
    arch = module_build_service.models.ModuleArch.query.get(1)
    num_contexts = 2 if contexts else 1
    for index in range(data_size):
        for context in range(num_contexts):
            build_one = ModuleBuild(
                name="nginx",
                stream="1",
                version=2 + index,
                state=BUILD_STATES["ready"],
                scratch=scratch,
                modulemd=read_staged_data("nginx_mmd"),
                koji_tag="scrmod-nginx-1.2" if scratch else "module-nginx-1.2",
                scmurl="git://pkgs.domain.local/modules/nginx"
                       "?#ba95886c7a443b36a9ce31abda1f9bef22f2f8c9",
                batch=2,
                # https://www.youtube.com/watch?v=iQGwrK_yDEg,
                owner="Moe Szyslak",
                time_submitted=datetime(2016, 9, 3, 11, 23, 20) + timedelta(minutes=(index * 10)),
                time_modified=datetime(2016, 9, 3, 11, 25, 32) + timedelta(minutes=(index * 10)),
                time_completed=datetime(2016, 9, 3, 11, 25, 32) + timedelta(minutes=(index * 10)),
                rebuild_strategy="changed-and-after",
            )
            build_one.arches.append(arch)

            if contexts:
                build_one.stream = str(index)
                nsvc = "{}:{}:{}:{}".format(
                    build_one.name,
                    build_one.stream,
                    build_one.version,
                    context
                )
                unique_hash = hashlib.sha1(nsvc.encode('utf-8')).hexdigest()
                build_one.build_context = unique_hash
                build_one.runtime_context = unique_hash
                build_one.ref_build_context = unique_hash
                combined_hashes = "{0}:{1}".format(unique_hash, unique_hash)
                build_one.context = hashlib.sha1(combined_hashes.encode("utf-8")).hexdigest()[:8]

            session.add(build_one)
            session.commit()
            build_one_component_release = get_rpm_release(build_one)

            component_one_build_one = ComponentBuild(
                package="nginx",
                scmurl="git://pkgs.domain.local/rpms/nginx?"
                       "#ga95886c8a443b36a9ce31abda1f9bed22f2f8c3",
                format="rpms",
                task_id=12312345 + index,
                state=koji.BUILD_STATES["COMPLETE"],
                nvr="nginx-1.10.1-2.{0}".format(build_one_component_release),
                batch=1,
                module_id=2 + index * 3,
                tagged=True,
                tagged_in_final=True,
            )

            component_two_build_one = ComponentBuild(
                package="module-build-macros",
                scmurl="/tmp/module_build_service-build-macrosWZUPeK/SRPMS/"
                       "module-build-macros-0.1-1.module_nginx_1_2.src.rpm",
                format="rpms",
                task_id=12312321 + index,
                state=koji.BUILD_STATES["COMPLETE"],
                nvr="module-build-macros-01-1.{0}".format(build_one_component_release),
                batch=2,
                module_id=2 + index * 3,
                tagged=True,
                tagged_in_final=True,
            )

        build_two = ModuleBuild(
            name="postgressql",
            stream="1",
            version=2 + index,
            state=BUILD_STATES["done"],
            scratch=scratch,
            modulemd=read_staged_data("testmodule"),
            koji_tag="scrmod-postgressql-1.2" if scratch else "module-postgressql-1.2",
            scmurl="git://pkgs.domain.local/modules/postgressql"
                   "?#aa95886c7a443b36a9ce31abda1f9bef22f2f8c9",
            batch=2,
            owner="some_user",
            time_submitted=datetime(2016, 9, 3, 12, 25, 33) + timedelta(minutes=(index * 10)),
            time_modified=datetime(2016, 9, 3, 12, 27, 19) + timedelta(minutes=(index * 10)),
            time_completed=datetime(2016, 9, 3, 11, 27, 19) + timedelta(minutes=(index * 10)),
            rebuild_strategy="changed-and-after",
        )
        build_two.arches.append(arch)

        session.add(build_two)
        session.commit()
        build_two_component_release = get_rpm_release(build_two)

        component_one_build_two = ComponentBuild(
            package="postgresql",
            scmurl="git://pkgs.domain.local/rpms/postgresql"
                   "?#dc95586c4a443b26a9ce38abda1f9bed22f2f8c3",
            format="rpms",
            task_id=2433433 + index,
            state=koji.BUILD_STATES["COMPLETE"],
            nvr="postgresql-9.5.3-4.{0}".format(build_two_component_release),
            batch=2,
            module_id=3 + index * 3,
            tagged=True,
            tagged_in_final=True,
        )

        component_two_build_two = ComponentBuild(
            package="module-build-macros",
            scmurl="/tmp/module_build_service-build-macrosWZUPeK/SRPMS/"
                   "module-build-macros-0.1-1.module_postgresql_1_2.src.rpm",
            format="rpms",
            task_id=47383993 + index,
            state=koji.BUILD_STATES["COMPLETE"],
            nvr="module-build-macros-01-1.{0}".format(build_two_component_release),
            batch=1,
            module_id=3 + index * 3,
        )

        build_three = ModuleBuild(
            name="testmodule",
            stream="4.3.43",
            version=6 + index,
            state=BUILD_STATES["wait"],
            scratch=scratch,
            modulemd=read_staged_data("testmodule"),
            koji_tag=None,
            scmurl="git://pkgs.domain.local/modules/testmodule"
                   "?#ca95886c7a443b36a9ce31abda1f9bef22f2f8c9",
            batch=0,
            owner="some_other_user",
            time_submitted=datetime(2016, 9, 3, 12, 28, 33) + timedelta(minutes=(index * 10)),
            time_modified=datetime(2016, 9, 3, 12, 28, 40) + timedelta(minutes=(index * 10)),
            time_completed=None,
            rebuild_strategy="changed-and-after",
        )
        session.add(build_three)
        session.commit()
        build_three_component_release = get_rpm_release(build_three)

        component_one_build_three = ComponentBuild(
            package="rubygem-rails",
            scmurl="git://pkgs.domain.local/rpms/rubygem-rails"
                   "?#dd55886c4a443b26a9ce38abda1f9bed22f2f8c3",
            format="rpms",
            task_id=2433433 + index,
            state=koji.BUILD_STATES["FAILED"],
            nvr="postgresql-9.5.3-4.{0}".format(build_three_component_release),
            batch=2,
            module_id=4 + index * 3,
        )

        component_two_build_three = ComponentBuild(
            package="module-build-macros",
            scmurl="/tmp/module_build_service-build-macrosWZUPeK/SRPMS/"
                   "module-build-macros-0.1-1.module_testmodule_1_2.src.rpm",
            format="rpms",
            task_id=47383993 + index,
            state=koji.BUILD_STATES["COMPLETE"],
            nvr="module-build-macros-01-1.{0}".format(build_three_component_release),
            batch=1,
            module_id=4 + index * 3,
            tagged=True,
            build_time_only=True,
        )

        session.add(component_one_build_one)
        session.add(component_two_build_one)
        session.add(component_one_build_two)
        session.add(component_two_build_two)
        session.add(component_one_build_three)
        session.add(component_two_build_three)
        session.commit()


def scheduler_init_data(tangerine_state=None, scratch=False):
    """ Creates a testmodule in the building state with all the components in the same batch
    """
    clean_database()

    current_dir = os.path.dirname(__file__)
    formatted_testmodule_yml_path = os.path.join(
        current_dir, "staged_data", "formatted_testmodule.yaml")
    mmd = load_mmd_file(formatted_testmodule_yml_path)
    mmd.get_rpm_component("tangerine").set_buildorder(0)

    platform_br = module_build_service.models.ModuleBuild.query.get(1)
    arch = module_build_service.models.ModuleArch.query.get(1)

    module_build = module_build_service.models.ModuleBuild(
        name="testmodule",
        stream="master",
        version=20170109091357,
        state=BUILD_STATES["build"],
        scratch=scratch,
        build_context="ac4de1c346dcf09ce77d38cd4e75094ec1c08eb0",
        runtime_context="ac4de1c346dcf09ce77d38cd4e75094ec1c08eb0",
        context="7c29193d",
        koji_tag="scrmod-testmodule-master-20170109091357-7c29193d"
                 if scratch
                 else "module-testmodule-master-20170109091357-7c29193d",
        scmurl="https://src.stg.fedoraproject.org/modules/testmodule.git?#ff1ea79",
        batch=3 if tangerine_state else 2,
        # https://www.youtube.com/watch?v=iOKymYVSaJE
        owner="Buzz Lightyear",
        time_submitted=datetime(2017, 2, 15, 16, 8, 18),
        time_modified=datetime(2017, 2, 15, 16, 19, 35),
        rebuild_strategy="changed-and-after",
        modulemd=mmd_to_str(mmd),
    )

    module_build.arches.append(arch)
    module_build.buildrequires.append(platform_br)
    build_one_component_release = get_rpm_release(module_build)

    module_build.component_builds.extend([
        module_build_service.models.ComponentBuild(
            package="perl-Tangerine",
            scmurl="https://src.fedoraproject.org/rpms/perl-Tangerine"
                   "?#4ceea43add2366d8b8c5a622a2fb563b625b9abf",
            format="rpms",
            task_id=90276227,
            state=koji.BUILD_STATES["COMPLETE"],
            nvr="perl-Tangerine-0.23-1.{0}".format(build_one_component_release),
            batch=2,
            ref="4ceea43add2366d8b8c5a622a2fb563b625b9abf",
            tagged=True,
            tagged_in_final=True,
        ),
        module_build_service.models.ComponentBuild(
            package="perl-List-Compare",
            scmurl="https://src.fedoraproject.org/rpms/perl-List-Compare"
                   "?#76f9d8c8e87eed0aab91034b01d3d5ff6bd5b4cb",
            format="rpms",
            task_id=90276228,
            state=koji.BUILD_STATES["COMPLETE"],
            nvr="perl-List-Compare-0.53-5.{0}".format(build_one_component_release),
            batch=2,
            ref="76f9d8c8e87eed0aab91034b01d3d5ff6bd5b4cb",
            tagged=True,
            tagged_in_final=True,
        ),
        module_build_service.models.ComponentBuild(
            package="tangerine",
            scmurl="https://src.fedoraproject.org/rpms/tangerine"
                   "?#fbed359411a1baa08d4a88e0d12d426fbf8f602c",
            format="rpms",
            batch=3,
            ref="fbed359411a1baa08d4a88e0d12d426fbf8f602c",
            state=tangerine_state,
            task_id=90276315 if tangerine_state else None,
            nvr="tangerine-0.22-3.{}".format(build_one_component_release)
            if tangerine_state
            else None,
            tagged=tangerine_state == koji.BUILD_STATES["COMPLETE"],
            tagged_in_final=tangerine_state == koji.BUILD_STATES["COMPLETE"],
        ),
        module_build_service.models.ComponentBuild(
            package="module-build-macros",
            scmurl="/tmp/module_build_service-build-macrosqr4AWH/SRPMS/module-build-"
                   "macros-0.1-1.module_testmodule_master_20170109091357.src.rpm",
            format="rpms",
            task_id=90276181,
            state=koji.BUILD_STATES["COMPLETE"],
            nvr="module-build-macros-0.1-1.{}".format(build_one_component_release),
            batch=1,
            tagged=True,
            build_time_only=True,
        ),
    ])

    with make_session(conf) as session:
        session.add(platform_br)
        session.add(module_build)
        session.commit()


def reuse_component_init_data():
    clean_database()

    current_dir = os.path.dirname(__file__)
    formatted_testmodule_yml_path = os.path.join(
        current_dir, "staged_data", "formatted_testmodule.yaml")
    mmd = load_mmd_file(formatted_testmodule_yml_path)

    platform_br = module_build_service.models.ModuleBuild.query.get(1)
    arch = module_build_service.models.ModuleArch.query.get(1)

    build_one = module_build_service.models.ModuleBuild(
        name="testmodule",
        stream="master",
        version=20170109091357,
        state=BUILD_STATES["ready"],
        ref_build_context="ac4de1c346dcf09ce77d38cd4e75094ec1c08eb0",
        runtime_context="ac4de1c346dcf09ce77d38cd4e75094ec1c08eb0",
        build_context="ac4de1c346dcf09ce77d38cd4e75094ec1c08eb1",
        context="78e4a6fd",
        koji_tag="module-testmodule-master-20170109091357-78e4a6fd",
        scmurl="https://src.stg.fedoraproject.org/modules/testmodule.git?#ff1ea79",
        batch=3,
        owner="Tom Brady",
        time_submitted=datetime(2017, 2, 15, 16, 8, 18),
        time_modified=datetime(2017, 2, 15, 16, 19, 35),
        time_completed=datetime(2017, 2, 15, 16, 19, 35),
        rebuild_strategy="changed-and-after",
    )

    build_one_component_release = get_rpm_release(build_one)

    mmd.set_version(int(build_one.version))
    xmd = mmd.get_xmd()
    xmd["mbs"]["scmurl"] = build_one.scmurl
    xmd["mbs"]["commit"] = "ff1ea79fc952143efeed1851aa0aa006559239ba"
    mmd.set_xmd(xmd)
    build_one.modulemd = mmd_to_str(mmd)
    build_one.arches.append(arch)
    build_one.buildrequires.append(platform_br)

    build_one.component_builds.extend([
        module_build_service.models.ComponentBuild(
            package="perl-Tangerine",
            scmurl="https://src.fedoraproject.org/rpms/perl-Tangerine"
                   "?#4ceea43add2366d8b8c5a622a2fb563b625b9abf",
            format="rpms",
            task_id=90276227,
            state=koji.BUILD_STATES["COMPLETE"],
            nvr="perl-Tangerine-0.23-1.{0}".format(build_one_component_release),
            batch=2,
            ref="4ceea43add2366d8b8c5a622a2fb563b625b9abf",
            tagged=True,
            tagged_in_final=True,
        ),
        module_build_service.models.ComponentBuild(
            package="perl-List-Compare",
            scmurl="https://src.fedoraproject.org/rpms/perl-List-Compare"
                   "?#76f9d8c8e87eed0aab91034b01d3d5ff6bd5b4cb",
            format="rpms",
            task_id=90276228,
            state=koji.BUILD_STATES["COMPLETE"],
            nvr="perl-List-Compare-0.53-5.{0}".format(build_one_component_release),
            batch=2,
            ref="76f9d8c8e87eed0aab91034b01d3d5ff6bd5b4cb",
            tagged=True,
            tagged_in_final=True,
        ),
        module_build_service.models.ComponentBuild(
            package="tangerine",
            scmurl="https://src.fedoraproject.org/rpms/tangerine"
                   "?#fbed359411a1baa08d4a88e0d12d426fbf8f602c",
            format="rpms",
            task_id=90276315,
            state=koji.BUILD_STATES["COMPLETE"],
            nvr="tangerine-0.22-3.{0}".format(build_one_component_release),
            batch=3,
            ref="fbed359411a1baa08d4a88e0d12d426fbf8f602c",
            tagged=True,
            tagged_in_final=True,
        ),
        module_build_service.models.ComponentBuild(
            package="module-build-macros",
            scmurl="/tmp/module_build_service-build-macrosqr4AWH/SRPMS/module-build-"
                   "macros-0.1-1.module_testmodule_master_20170109091357.src.rpm",
            format="rpms",
            task_id=90276181,
            state=koji.BUILD_STATES["COMPLETE"],
            nvr="module-build-macros-0.1-1.{0}".format(build_one_component_release),
            batch=1,
            tagged=True,
            build_time_only=True,
        ),
    ])

    build_two = module_build_service.models.ModuleBuild(
        name="testmodule",
        stream="master",
        version=20170219191323,
        state=BUILD_STATES["build"],
        ref_build_context="ac4de1c346dcf09ce77d38cd4e75094ec1c08eb0",
        runtime_context="ac4de1c346dcf09ce77d38cd4e75094ec1c08eb0",
        build_context="ac4de1c346dcf09ce77d38cd4e75094ec1c08eb1",
        context="c40c156c",
        koji_tag="module-testmodule-master-20170219191323-c40c156c",
        scmurl="https://src.stg.fedoraproject.org/modules/testmodule.git?#55f4a0a",
        batch=1,
        owner="Tom Brady",
        time_submitted=datetime(2017, 2, 19, 16, 8, 18),
        time_modified=datetime(2017, 2, 19, 16, 8, 18),
        rebuild_strategy="changed-and-after",
    )

    build_two_component_release = get_rpm_release(build_two)

    mmd.set_version(int(build_one.version))
    xmd = mmd.get_xmd()
    xmd["mbs"]["scmurl"] = build_one.scmurl
    xmd["mbs"]["commit"] = "55f4a0a2e6cc255c88712a905157ab39315b8fd8"
    mmd.set_xmd(xmd)
    build_two.modulemd = mmd_to_str(mmd)
    build_two.arches.append(arch)
    build_two.buildrequires.append(platform_br)

    build_two.component_builds.extend([
        module_build_service.models.ComponentBuild(
            package="perl-Tangerine",
            scmurl="https://src.fedoraproject.org/rpms/perl-Tangerine"
                   "?#4ceea43add2366d8b8c5a622a2fb563b625b9abf",
            format="rpms",
            batch=2,
            ref="4ceea43add2366d8b8c5a622a2fb563b625b9abf",
        ),
        module_build_service.models.ComponentBuild(
            package="perl-List-Compare",
            scmurl="https://src.fedoraproject.org/rpms/perl-List-Compare"
                   "?#76f9d8c8e87eed0aab91034b01d3d5ff6bd5b4cb",
            format="rpms",
            batch=2,
            ref="76f9d8c8e87eed0aab91034b01d3d5ff6bd5b4cb",
        ),
        module_build_service.models.ComponentBuild(
            package="tangerine",
            scmurl="https://src.fedoraproject.org/rpms/tangerine"
                   "?#fbed359411a1baa08d4a88e0d12d426fbf8f602c",
            format="rpms",
            batch=3,
            ref="fbed359411a1baa08d4a88e0d12d426fbf8f602c",
        ),
        module_build_service.models.ComponentBuild(
            package="module-build-macros",
            scmurl="/tmp/module_build_service-build-macrosqr4AWH/SRPMS/module-build-"
                   "macros-0.1-1.module_testmodule_master_20170219191323.src.rpm",
            format="rpms",
            task_id=90276186,
            state=koji.BUILD_STATES["COMPLETE"],
            nvr="module-build-macros-0.1-1.{0}".format(build_two_component_release),
            batch=1,
            tagged=True,
            build_time_only=True,
        ),
    ])

    with make_session(conf) as session:
        session.add(platform_br)
        session.add(build_one)
        session.add(build_two)
        session.commit()


def reuse_shared_userspace_init_data():
    clean_database()

    with make_session(conf) as session:
        # Create shared-userspace-570, state is COMPLETE, all components
        # are properly built.
        current_dir = os.path.dirname(__file__)
        formatted_testmodule_yml_path = os.path.join(
            current_dir, "staged_data", "shared-userspace-570.yaml")
        mmd = load_mmd_file(formatted_testmodule_yml_path)

        module_build = module_build_service.models.ModuleBuild(
            name=mmd.get_module_name(),
            stream=mmd.get_stream_name(),
            version=mmd.get_version(),
            build_context="e046b867a400a06a3571f3c71142d497895fefbe",
            runtime_context="50dd3eb5dde600d072e45d4120e1548ce66bc94a",
            state=BUILD_STATES["ready"],
            modulemd=mmd_to_str(mmd),
            koji_tag="module-shared-userspace-f26-20170601141014-75f92abb",
            scmurl="https://src.stg.fedoraproject.org/modules/testmodule.git?#7fea453",
            batch=16,
            owner="Tom Brady",
            time_submitted=datetime(2017, 2, 15, 16, 8, 18),
            time_modified=datetime(2017, 2, 15, 16, 19, 35),
            time_completed=datetime(2017, 2, 15, 16, 19, 35),
            rebuild_strategy="changed-and-after",
        )

        components = [
            mmd.get_rpm_component(rpm)
            for rpm in mmd.get_rpm_component_names()
        ]
        components.sort(key=lambda x: x.get_buildorder())
        previous_buildorder = None
        batch = 1
        for pkg in components:
            # Increment the batch number when buildorder increases.
            if previous_buildorder != pkg.get_buildorder():
                previous_buildorder = pkg.get_buildorder()
                batch += 1

            pkgref = mmd.get_xmd()["mbs"]["rpms"][pkg.get_name()]["ref"]
            full_url = pkg.get_repository() + "?#" + pkgref

            module_build.component_builds.append(
                module_build_service.models.ComponentBuild(
                    package=pkg.get_name(),
                    format="rpms",
                    scmurl=full_url,
                    batch=batch,
                    ref=pkgref,
                    state=1,
                    tagged=True,
                    tagged_in_final=True,
                )
            )

        session.add(module_build)
        session.commit()

        # Create shared-userspace-577, state is WAIT, no component built
        formatted_testmodule_yml_path = os.path.join(
            current_dir, "staged_data", "shared-userspace-577.yaml"
        )
        mmd2 = load_mmd_file(formatted_testmodule_yml_path)

        module_build = module_build_service.models.ModuleBuild(
            name=mmd2.get_module_name(),
            stream=mmd2.get_stream_name(),
            version=mmd2.get_version(),
            build_context="e046b867a400a06a3571f3c71142d497895fefbe",
            runtime_context="50dd3eb5dde600d072e45d4120e1548ce66bc94a",
            state=BUILD_STATES["done"],
            modulemd=mmd_to_str(mmd2),
            koji_tag="module-shared-userspace-f26-20170605091544-75f92abb",
            scmurl="https://src.stg.fedoraproject.org/modules/testmodule.git?#7fea453",
            batch=0,
            owner="Tom Brady",
            time_submitted=datetime(2017, 2, 15, 16, 8, 18),
            time_modified=datetime(2017, 2, 15, 16, 19, 35),
            time_completed=datetime(2017, 2, 15, 16, 19, 35),
            rebuild_strategy="changed-and-after",
        )

        components2 = [
            mmd2.get_rpm_component(rpm)
            for rpm in mmd2.get_rpm_component_names()
        ]
        # Store components to database in different order than for 570 to
        # reproduce the reusing issue.
        components2.sort(key=lambda x: len(x.get_name()))
        components2.sort(key=lambda x: x.get_buildorder())
        previous_buildorder = None
        batch = 1
        for pkg in components2:
            # Increment the batch number when buildorder increases.
            if previous_buildorder != pkg.get_buildorder():
                previous_buildorder = pkg.get_buildorder()
                batch += 1

            pkgref = mmd2.get_xmd()["mbs"]["rpms"][pkg.get_name()]["ref"]
            full_url = pkg.get_repository() + "?#" + pkgref

            module_build.component_builds.append(
                module_build_service.models.ComponentBuild(
                    package=pkg.get_name(), format="rpms", scmurl=full_url, batch=batch, ref=pkgref)
            )

        session.add(module_build)
        session.commit()


def make_module(
    nsvc,
    requires_list=None,
    build_requires_list=None,
    base_module=None,
    filtered_rpms=None,
    xmd=None,
    store_to_db=True,
    virtual_streams=None,
    arches=None,
):
    """
    Creates new models.ModuleBuild defined by `nsvc` string with requires
    and buildrequires set according to ``requires_list`` and ``build_requires_list``.

    :param str nsvc: name:stream:version:context of a module.
    :param list_of_dicts requires_list: List of dictionaries defining the
        requires in the mmd requires field format.
    :param list_of_dicts build_requires_list: List of dictionaries defining the
        build_requires_list in the mmd build_requires_list field format.
    :param filtered_rpms: list of filtered RPMs which are added to filter
        section in module metadata.
    :type filtered_rpms: list[str]
    :param dict xmd: a mapping representing XMD section in module metadata. A
        custom xmd could be passed for testing a particular scenario and some
        default key/value pairs are added if not present.
    :param bool store_to_db: whether to store created module metadata to the
        database.
    :param list virtual_streams: List of virtual streams provided by this module.
    :param list arches: List of architectures this module is built against.
        If set to None, ["x86_64"] is used as a default.
    :return: New Module Build if set to store module metadata to database,
        otherwise the module metadata is returned.
    :rtype: ModuleBuild or Modulemd.Module
    """
    name, stream, version, context = nsvc.split(":")
    mmd = Modulemd.ModuleStreamV2.new(name, stream)
    mmd.set_version(int(version))
    mmd.set_context(context)
    mmd.set_summary("foo")
    # Test unicode in mmd.
    mmd.set_description(u"foo \u2019s")
    mmd.add_module_license("GPL")

    if filtered_rpms:
        for rpm in filtered_rpms:
            mmd.add_rpm_filter(rpm)

    if requires_list is not None and build_requires_list is not None:
        if not isinstance(requires_list, list):
            requires_list = [requires_list]
        if not isinstance(build_requires_list, list):
            build_requires_list = [build_requires_list]

        for requires, build_requires in zip(requires_list, build_requires_list):
            deps = Modulemd.Dependencies()
            for req_name, req_streams in requires.items():
                if req_streams == []:
                    deps.set_empty_runtime_dependencies_for_module(req_name)
                else:
                    for req_stream in req_streams:
                        deps.add_runtime_stream(req_name, req_stream)

            for req_name, req_streams in build_requires.items():
                if req_streams == []:
                    deps.set_empty_buildtime_dependencies_for_module(req_name)
                else:
                    for req_stream in req_streams:
                        deps.add_buildtime_stream(req_name, req_stream)

            mmd.add_dependencies(deps)

    # Caller could pass whole xmd including mbs, but if something is missing,
    # default values are given here.
    xmd = xmd or {"mbs": {}}
    xmd_mbs = xmd["mbs"]
    if "buildrequires" not in xmd_mbs:
        xmd_mbs["buildrequires"] = {}
    if "requires" not in xmd_mbs:
        xmd_mbs["requires"] = {}
    if "commit" not in xmd_mbs:
        xmd_mbs["commit"] = "ref_%s" % context
    if "mse" not in xmd_mbs:
        xmd_mbs["mse"] = "true"

    if virtual_streams:
        xmd_mbs["virtual_streams"] = virtual_streams

    mmd.set_xmd(xmd)

    if not store_to_db:
        return mmd

    module_build = ModuleBuild(
        name=name,
        stream=stream,
        stream_version=ModuleBuild.get_stream_version(stream),
        version=version,
        context=context,
        state=BUILD_STATES["ready"],
        scmurl="https://src.stg.fedoraproject.org/modules/unused.git?#ff1ea79",
        batch=1,
        owner="Tom Brady",
        time_submitted=datetime(2017, 2, 15, 16, 8, 18),
        time_modified=datetime(2017, 2, 15, 16, 19, 35),
        rebuild_strategy="changed-and-after",
        build_context=context,
        runtime_context=context,
        modulemd=mmd_to_str(mmd),
        koji_tag=xmd["mbs"]["koji_tag"] if "koji_tag" in xmd["mbs"] else None,
    )
    if base_module:
        module_build.buildrequires.append(base_module)
    db.session.add(module_build)
    db.session.commit()

    if virtual_streams:
        for virtual_stream in virtual_streams:
            vs_obj = db.session.query(VirtualStream).filter_by(name=virtual_stream).first()
            if not vs_obj:
                vs_obj = VirtualStream(name=virtual_stream)
                db.session.add(vs_obj)
                db.session.commit()

            if vs_obj not in module_build.virtual_streams:
                module_build.virtual_streams.append(vs_obj)
                db.session.commit()

    if arches is None:
        arches = ["x86_64"]
    for arch in arches:
        arch_obj = db.session.query(module_build_service.models.ModuleArch).filter_by(
            name=arch).first()
        if not arch_obj:
            arch_obj = module_build_service.models.ModuleArch(name=arch)
            db.session.add(arch_obj)
            db.session.commit()

        if arch_obj not in module_build.arches:
            module_build.arches.append(arch_obj)
            db.session.commit()

    return module_build



# -*- coding: utf-8 -*-
# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Ralph Bean <rbean@redhat.com>
#            Matt Prahl <mprahl@redhat.com>

"""Generic messaging functions."""

import re
import pkg_resources

try:
    from inspect import signature
except ImportError:
    from funcsigs import signature

from module_build_service import log


class IgnoreMessage(Exception):
    pass


class BaseMessage(object):
    def __init__(self, msg_id):
        """
        A base class to abstract messages from different backends
        :param msg_id: the id of the msg (e.g. 2016-SomeGUID)
        """
        self.msg_id = msg_id

        # Moksha calls `consumer.validate` on messages that it receives, and
        # even though we have validation turned off in the config there's still
        # a step that tries to access `msg['body']`, `msg['topic']` and
        # `msg.get('topic')`.
        # These are here just so that the `validate` method won't raise an
        # exception when we push our fake messages through.
        # Note that, our fake message pushing has worked for a while... but the
        # *latest* version of fedmsg has some code that exercises the bug.  I
        # didn't hit this until I went to test in jenkins.
        self.body = {}
        self.topic = None

    def __repr__(self):
        init_sig = signature(self.__init__)

        args_strs = (
            "{}={!r}".format(name, getattr(self, name))
            if param.default != param.empty
            else repr(getattr(self, name))
            for name, param in init_sig.parameters.items()
        )

        return "{}({})".format(type(self).__name__, ", ".join(args_strs))

    def __getitem__(self, key):
        """ Used to trick moksha into thinking we are a dict. """
        return getattr(self, key)

    def __setitem__(self, key, value):
        """ Used to trick moksha into thinking we are a dict. """
        return setattr(self, key, value)

    def get(self, key, value=None):
        """ Used to trick moksha into thinking we are a dict. """
        return getattr(self, key, value)

    def __json__(self):
        return dict(msg_id=self.msg_id, topic=self.topic, body=self.body)


class MessageParser(object):
    def parse(self, msg):
        raise NotImplementedError()


class FedmsgMessageParser(MessageParser):
    def parse(self, msg):
        """
        Takes a fedmsg topic and message and converts it to a message object
        :param msg: the message contents from the fedmsg message
        :return: an object of BaseMessage descent if the message is a type
        that the app looks for, otherwise None is returned
        """
        if "body" in msg:
            msg = msg["body"]
        topic = msg["topic"]
        topic_categories = _messaging_backends["fedmsg"]["services"]
        categories_re = "|".join(map(re.escape, topic_categories))
        regex_pattern = re.compile(
            r"(?P<category>" + categories_re + r")"
            r"(?:(?:\.)(?P<object>build|repo|module|decision))?"
            r"(?:(?:\.)(?P<subobject>state|build))?"
            r"(?:\.)(?P<event>change|done|end|tag|update)$"
        )
        regex_results = re.search(regex_pattern, topic)

        if regex_results:
            category = regex_results.group("category")
            object = regex_results.group("object")
            subobject = regex_results.group("subobject")
            event = regex_results.group("event")

            msg_id = msg.get("msg_id")
            msg_inner_msg = msg.get("msg")

            # If there isn't a msg dict in msg then this message can be skipped
            if not msg_inner_msg:
                log.debug(
                    "Skipping message without any content with the " 'topic "{0}"'.format(topic))
                return None

            msg_obj = None

            # Ignore all messages from the secondary koji instances.
            if category == "buildsys":
                instance = msg_inner_msg.get("instance", "primary")
                if instance != "primary":
                    log.debug("Ignoring message from %r koji hub." % instance)
                    return

            if (
                category == "buildsys"
                and object == "build"
                and subobject == "state"
                and event == "change"
            ):
                build_id = msg_inner_msg.get("build_id")
                task_id = msg_inner_msg.get("task_id")
                build_new_state = msg_inner_msg.get("new")
                build_name = msg_inner_msg.get("name")
                build_version = msg_inner_msg.get("version")
                build_release = msg_inner_msg.get("release")

                msg_obj = KojiBuildChange(
                    msg_id,
                    build_id,
                    task_id,
                    build_new_state,
                    build_name,
                    build_version,
                    build_release,
                )

            elif (
                category == "buildsys"
                and object == "repo"
                and subobject is None
                and event == "done"
            ):
                repo_tag = msg_inner_msg.get("tag")
                msg_obj = KojiRepoChange(msg_id, repo_tag)

            elif category == "buildsys" and event == "tag":
                tag = msg_inner_msg.get("tag")
                name = msg_inner_msg.get("name")
                version = msg_inner_msg.get("version")
                release = msg_inner_msg.get("release")
                nvr = None
                if name and version and release:
                    nvr = "-".join((name, version, release))
                msg_obj = KojiTagChange(msg_id, tag, name, nvr)

            elif (
                category == "mbs"
                and object == "module"
                and subobject == "state"
                and event == "change"
            ):
                msg_obj = MBSModule(msg_id, msg_inner_msg.get("id"), msg_inner_msg.get("state"))

            elif (
                category == "greenwave"
                and object == "decision"
                and subobject is None
                and event == "update"
            ):
                msg_obj = GreenwaveDecisionUpdate(
                    msg_id=msg_id,
                    decision_context=msg_inner_msg.get("decision_context"),
                    policies_satisfied=msg_inner_msg.get("policies_satisfied"),
                    subject_identifier=msg_inner_msg.get("subject_identifier"),
                )

            # If the message matched the regex and is important to the app,
            # it will be returned
            if msg_obj:
                return msg_obj

        return None


class KojiBuildChange(BaseMessage):
    """ A class that inherits from BaseMessage to provide a message
    object for a build's info (in fedmsg this replaces the msg dictionary)
    :param msg_id: the id of the msg (e.g. 2016-SomeGUID)
    :param build_id: the id of the build (e.g. 264382)
    :param build_new_state: the new build state, this is currently a Koji
    integer
    :param build_name: the name of what is being built
    (e.g. golang-googlecode-tools)
    :param build_version: the version of the build (e.g. 6.06.06)
    :param build_release: the release of the build (e.g. 4.fc25)
    :param module_build_id: the optional id of the module_build in the database
    :param state_reason: the optional reason as to why the state changed
    """

    def __init__(
        self,
        msg_id,
        build_id,
        task_id,
        build_new_state,
        build_name,
        build_version,
        build_release,
        module_build_id=None,
        state_reason=None,
    ):
        if task_id is None:
            raise IgnoreMessage("KojiBuildChange with a null task_id is invalid.")
        super(KojiBuildChange, self).__init__(msg_id)
        self.build_id = build_id
        self.task_id = task_id
        self.build_new_state = build_new_state
        self.build_name = build_name
        self.build_version = build_version
        self.build_release = build_release
        self.module_build_id = module_build_id
        self.state_reason = state_reason


class KojiTagChange(BaseMessage):
    """
    A class that inherits from BaseMessage to provide a message
    object for a buildsys.tag info (in fedmsg this replaces the msg dictionary)
    :param tag: the name of tag (e.g. module-123456789-build)
    :param artifact: the name of tagged artifact (e.g. module-build-macros)
    :param nvr: the nvr of the tagged artifact
    """

    def __init__(self, msg_id, tag, artifact, nvr):
        super(KojiTagChange, self).__init__(msg_id)
        self.tag = tag
        self.artifact = artifact
        self.nvr = nvr


class KojiRepoChange(BaseMessage):
    """ A class that inherits from BaseMessage to provide a message
    object for a repo's info (in fedmsg this replaces the msg dictionary)
    :param msg_id: the id of the msg (e.g. 2016-SomeGUID)
    :param repo_tag: the repo's tag (e.g. SHADOWBUILD-f25-build)
    """

    def __init__(self, msg_id, repo_tag):
        super(KojiRepoChange, self).__init__(msg_id)
        self.repo_tag = repo_tag


class MBSModule(BaseMessage):
    """ A class that inherits from BaseMessage to provide a message
    object for a module event generated by module_build_service
    :param msg_id: the id of the msg (e.g. 2016-SomeGUID)
    :param module_build_id: the id of the module build
    :param module_build_state: the state of the module build
    """

    def __init__(self, msg_id, module_build_id, module_build_state):
        super(MBSModule, self).__init__(msg_id)
        self.module_build_id = module_build_id
        self.module_build_state = module_build_state


class GreenwaveDecisionUpdate(BaseMessage):
    """A class representing message send to topic greenwave.decision.update"""

    def __init__(self, msg_id, decision_context, policies_satisfied, subject_identifier):
        super(GreenwaveDecisionUpdate, self).__init__(msg_id)
        self.decision_context = decision_context
        self.policies_satisfied = policies_satisfied
        self.subject_identifier = subject_identifier


def publish(topic, msg, conf, service):
    """
    Publish a single message to a given backend, and return
    :param topic: the topic of the message (e.g. module.state.change)
    :param msg: the message contents of the message (typically JSON)
    :param conf: a Config object from the class in config.py
    :param service: the system that is publishing the message (e.g. mbs)
    :return:
    """
    try:
        handler = _messaging_backends[conf.messaging]["publish"]
    except KeyError:
        raise KeyError(
            "No messaging backend found for %r in %r" % (conf.messaging, _messaging_backends.keys())
        )

    from module_build_service.monitor import (
        messaging_tx_to_send_counter,
        messaging_tx_sent_ok_counter,
        messaging_tx_failed_counter,
    )

    messaging_tx_to_send_counter.inc()
    try:
        rv = handler(topic, msg, conf, service)
        messaging_tx_sent_ok_counter.inc()
        return rv
    except Exception:
        messaging_tx_failed_counter.inc()
        raise


def _fedmsg_publish(topic, msg, conf, service):
    # fedmsg doesn't really need access to conf, however other backends do
    import fedmsg

    return fedmsg.publish(topic, msg=msg, modname=service)


# A counter used for in-memory messages.
_in_memory_msg_id = 0
_initial_messages = []


def _in_memory_publish(topic, msg, conf, service):
    """ Puts the message into the in memory work queue. """
    # Increment the message ID.
    global _in_memory_msg_id
    _in_memory_msg_id += 1

    # Create fake fedmsg from the message so we can reuse
    # the BaseMessage.from_fedmsg code to get the particular BaseMessage
    # class instance.
    wrapped_msg = FedmsgMessageParser().parse({
        "msg_id": str(_in_memory_msg_id),
        "topic": service + "." + topic,
        "msg": msg
    })

    # Put the message to queue.
    from module_build_service.scheduler.consumer import work_queue_put

    try:
        work_queue_put(wrapped_msg)
    except ValueError as e:
        log.warning("No MBSConsumer found.  Shutting down?  %r" % e)
    except AttributeError:
        # In the event that `moksha.hub._hub` hasn't yet been initialized, we
        # need to store messages on the side until it becomes available.
        # As a last-ditch effort, try to hang initial messages in the config.
        log.warning("Hub not initialized.  Queueing on the side.")
        _initial_messages.append(wrapped_msg)


_fedmsg_backend = {
    "publish": _fedmsg_publish,
    "services": ["buildsys", "mbs", "greenwave"],
    "parser": FedmsgMessageParser(),
    "topic_suffix": ".",
}
_in_memory_backend = {
    "publish": _in_memory_publish,
    "services": [],
    "parser": FedmsgMessageParser(),  # re-used.  :)
    "topic_suffix": ".",
}


_messaging_backends = {}
for entrypoint in pkg_resources.iter_entry_points("mbs.messaging_backends"):
    _messaging_backends[entrypoint.name] = ep = entrypoint.load()
    required = ["publish", "services", "parser", "topic_suffix"]
    if any([key not in ep for key in required]):
        raise ValueError("messaging backend %r is malformed: %r" % (entrypoint.name, ep))

if not _messaging_backends:
    raise ValueError("No messaging plugins are installed or available.")



# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

import pytest

from module_build_service.utils.general import load_mmd_file, mmd_to_str


BASE_DIR = os.path.dirname(__file__)
STAGED_DATA_DIR = os.path.join(BASE_DIR, "staged_data")

_mmd = load_mmd_file(os.path.join(STAGED_DATA_DIR, "platform.yaml"))
PLATFORM_MODULEMD = mmd_to_str(_mmd)

_mmd2 = load_mmd_file(os.path.join(STAGED_DATA_DIR, "formatted_testmodule.yaml"))
TESTMODULE_MODULEMD = mmd_to_str(_mmd2)

_mmd3 = load_mmd_file(os.path.join(STAGED_DATA_DIR, "formatted_testmodule.yaml"))
_mmd3.set_context("c2c572ed")
TESTMODULE_MODULEMD_SECOND_CONTEXT = mmd_to_str(_mmd3)


@pytest.fixture()
def testmodule_mmd_9c690d0e():
    return TESTMODULE_MODULEMD


@pytest.fixture()
def testmodule_mmd_c2c572ed():
    return TESTMODULE_MODULEMD_SECOND_CONTEXT


@pytest.fixture()
def formatted_testmodule_mmd():
    return _mmd2


@pytest.fixture()
def platform_mmd():
    return PLATFORM_MODULEMD



# Copyright (c) 2019  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Valerij Maljulin <vmaljuli@redhat.com>


import json
from mock import patch, Mock
import pytest
from module_build_service.utils.greenwave import greenwave
from tests import make_module


class TestGreenwaveQuery():
    @patch("module_build_service.utils.greenwave.requests")
    def test_greenwave_query_decision(self, mock_requests):
        resp_status = 200
        resp_content = {
            "applicable_policies": ["osci_compose_modules"],
            "policies_satisfied": True,
            "satisfied_requirements": [
                {
                    "result_id": 7336633,
                    "testcase": "test-ci.test-module.tier1",
                    "type": "test-result-passed"
                },
                {
                    "result_id": 7336650,
                    "testcase": "test-ci.test-module.tier2",
                    "type": "test-result-passed"
                }
            ],
            "summary": "All required tests passed",
            "unsatisfied_requirements": []
        }
        response = Mock()
        response.json.return_value = resp_content
        response.status_code = resp_status
        mock_requests.post.return_value = response

        fake_build = make_module("pkg:0.1:1:c1", requires_list={"platform": "el8"})

        got_response = greenwave.query_decision(fake_build, prod_version="xxxx-8")

        assert got_response == resp_content
        assert json.loads(mock_requests.post.call_args_list[0][1]["data"]) == {
            "decision_context": "test_dec_context",
            "product_version": "xxxx-8", "subject_type": "some-module",
            "subject_identifier": "pkg-0.1-1.c1"}
        assert mock_requests.post.call_args_list[0][1]["headers"] == {
            "Content-Type": "application/json"}
        assert mock_requests.post.call_args_list[0][1]["url"] == \
            "https://greenwave.example.local/api/v1.0/decision"

    @pytest.mark.parametrize("return_all", (False, True))
    @patch("module_build_service.utils.greenwave.requests")
    def test_greenwave_query_policies(self, mock_requests, return_all):
        resp_status = 200
        resp_content = {
            "policies": [
                {
                    "decision_context": "test_dec_context",
                    "product_versions": ["ver1", "ver3"],
                    "rules": [],
                    "subject_type": "some-module"
                },
                {
                    "decision_context": "test_dec_context",
                    "product_versions": ["ver1", "ver2"],
                    "rules": [],
                    "subject_type": "some-module"
                },
                {
                    "decision_context": "decision_context_2",
                    "product_versions": ["ver4"],
                    "rules": [],
                    "subject_type": "subject_type_2"
                }
            ]
        }
        selected_policies = {"policies": resp_content["policies"][:-1]}

        response = Mock()
        response.json.return_value = resp_content
        response.status_code = resp_status
        mock_requests.get.return_value = response

        got_response = greenwave.query_policies(return_all)

        if return_all:
            assert got_response == resp_content
        else:
            assert got_response == selected_policies
        assert mock_requests.get.call_args_list[0][1]["url"] == \
            "https://greenwave.example.local/api/v1.0/policies"

    @patch("module_build_service.utils.greenwave.requests")
    def test_greenwave_get_product_versions(self, mock_requests):
        resp_status = 200
        resp_content = {
            "policies": [
                {
                    "decision_context": "test_dec_context",
                    "product_versions": ["ver1", "ver3"],
                    "rules": [],
                    "subject_type": "some-module"
                },
                {
                    "decision_context": "test_dec_context",
                    "product_versions": ["ver1", "ver2"],
                    "rules": [],
                    "subject_type": "some-module"
                },
                {
                    "decision_context": "decision_context_2",
                    "product_versions": ["ver4"],
                    "rules": [],
                    "subject_type": "subject_type_2"
                }
            ]
        }
        expected_versions = {"ver1", "ver2", "ver3"}

        response = Mock()
        response.json.return_value = resp_content
        response.status_code = resp_status
        mock_requests.get.return_value = response

        versions_set = greenwave.get_product_versions()

        assert versions_set == expected_versions
        assert mock_requests.get.call_args_list[0][1]["url"] == \
            "https://greenwave.example.local/api/v1.0/policies"

    @pytest.mark.parametrize("policies_satisfied", (True, False))
    @patch("module_build_service.utils.greenwave.requests")
    def test_greenwave_check_gating(self, mock_requests, policies_satisfied):
        resp_status = 200
        policies_content = {
            "policies": [
                {
                    "decision_context": "test_dec_context",
                    "product_versions": ["ver1", "ver3"],
                    "rules": [],
                    "subject_type": "some-module"
                }
            ]
        }

        responses = [Mock() for i in range(3)]
        for r in responses:
            r.status_code = resp_status
        responses[0].json.return_value = policies_content
        responses[1].json.return_value = {"policies_satisfied": False}
        responses[2].json.return_value = {"policies_satisfied": policies_satisfied}
        mock_requests.get.return_value = responses[0]
        mock_requests.post.side_effect = responses[1:]

        fake_build = make_module("pkg:0.1:1:c1", requires_list={"platform": "el8"})
        result = greenwave.check_gating(fake_build)

        assert result == policies_satisfied



# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Ralph Bean <rbean@redhat.com>

import mock
from mock import patch
import module_build_service.messaging
import module_build_service.scheduler.handlers.modules
import os
import koji
import pytest
from tests import conf, db, app, scheduler_init_data
import module_build_service.resolver
from module_build_service import build_logs, Modulemd
from module_build_service.utils.general import load_mmd_file
from module_build_service.models import ComponentBuild, ModuleBuild

base_dir = os.path.dirname(os.path.dirname(__file__))


class TestModuleWait:
    def setup_method(self, test_method):
        self.config = conf
        self.session = mock.Mock()
        self.fn = module_build_service.scheduler.handlers.modules.wait

    def teardown_method(self, test_method):
        try:
            path = build_logs.path(1)
            os.remove(path)
        except Exception:
            pass

    @patch("module_build_service.builder.GenericBuilder.create_from_module")
    @patch("module_build_service.models.ModuleBuild.from_module_event")
    def test_init_basic(self, from_module_event, create_builder):
        builder = mock.Mock()
        builder.get_disttag_srpm.return_value = "some srpm disttag"
        builder.build.return_value = 1234, 1, "", None
        builder.module_build_tag = {"name": "some-tag-build"}
        create_builder.return_value = builder
        mocked_module_build = mock.Mock()
        mocked_module_build.name = "foo"
        mocked_module_build.stream = "stream"
        mocked_module_build.version = "1"
        mocked_module_build.context = "1234567"
        mocked_module_build.state = 1
        mocked_module_build.id = 1
        mocked_module_build.json.return_value = {
            "name": "foo",
            "stream": "1",
            "version": 1,
            "state": "some state",
            "id": 1,
        }

        formatted_testmodule_yml_path = os.path.join(
            base_dir, "staged_data", "formatted_testmodule.yaml")
        mmd = load_mmd_file(formatted_testmodule_yml_path)
        mocked_module_build.id = 1
        mocked_module_build.mmd.return_value = mmd
        mocked_module_build.component_builds = []

        from_module_event.return_value = mocked_module_build

        msg = module_build_service.messaging.MBSModule(
            msg_id=None, module_build_id=1, module_build_state="some state")
        with patch.object(module_build_service.resolver, "system_resolver"):
            self.fn(config=self.config, session=self.session, msg=msg)

    @patch(
        "module_build_service.builder.GenericBuilder.default_buildroot_groups",
        return_value={"build": [], "srpm-build": []},
    )
    @patch("module_build_service.builder.GenericBuilder.create_from_module")
    @patch("module_build_service.resolver.DBResolver")
    @patch("module_build_service.resolver.GenericResolver")
    def test_new_repo_called_when_macros_reused(
        self, generic_resolver, resolver, create_builder, dbg
    ):
        """
        Test that newRepo is called when module-build-macros build is reused.
        """
        with app.app_context():
            scheduler_init_data()
            koji_session = mock.MagicMock()
            koji_session.newRepo.return_value = 123456

            builder = mock.MagicMock()
            builder.koji_session = koji_session
            builder.module_build_tag = {"name": "module-123-build"}
            builder.get_disttag_srpm.return_value = "some srpm disttag"
            builder.build.return_value = (
                1234,
                koji.BUILD_STATES["COMPLETE"],
                "",
                "module-build-macros-1-1",
            )
            create_builder.return_value = builder

            resolver = mock.MagicMock()
            resolver.backend = "db"
            resolver.get_module_tag.return_value = "module-testmodule-master-20170109091357"

            with patch.object(module_build_service.resolver, "system_resolver", new=resolver):
                msg = module_build_service.messaging.MBSModule(
                    msg_id=None, module_build_id=2, module_build_state="some state")
                module_build_service.scheduler.handlers.modules.wait(
                    config=conf, session=db.session, msg=msg)
                koji_session.newRepo.assert_called_once_with("module-123-build")

            # When module-build-macros is reused, it still has to appear only
            # once in database.
            builds_count = (
                db.session.query(ComponentBuild)
                .filter_by(package="module-build-macros", module_id=2)
                .count()
            )
            assert builds_count == 1

    @patch(
        "module_build_service.builder.GenericBuilder.default_buildroot_groups",
        return_value={"build": [], "srpm-build": []},
    )
    @patch("module_build_service.builder.GenericBuilder.create_from_module")
    @patch("module_build_service.resolver.DBResolver")
    @patch("module_build_service.resolver.GenericResolver")
    def test_new_repo_not_called_when_macros_not_reused(
        self, generic_resolver, resolver, create_builder, dbg
    ):
        """
        Test that newRepo is called everytime for module-build-macros
        """
        with app.app_context():
            scheduler_init_data()
            koji_session = mock.MagicMock()
            koji_session.newRepo.return_value = 123456

            builder = mock.MagicMock()
            builder.koji_session = koji_session
            builder.module_build_tag = {"name": "module-123-build"}
            builder.get_disttag_srpm.return_value = "some srpm disttag"
            builder.build.return_value = (
                1234,
                koji.BUILD_STATES["BUILDING"],
                "",
                "module-build-macros-1-1",
            )
            create_builder.return_value = builder

            resolver = mock.MagicMock()
            resolver.backend = "db"
            resolver.get_module_tag.return_value = "module-testmodule-master-20170109091357"

            with patch.object(module_build_service.resolver, "system_resolver", new=resolver):
                msg = module_build_service.messaging.MBSModule(
                    msg_id=None, module_build_id=2, module_build_state="some state")
                module_build_service.scheduler.handlers.modules.wait(
                    config=conf, session=db.session, msg=msg)
                assert koji_session.newRepo.called

    @patch(
        "module_build_service.builder.GenericBuilder.default_buildroot_groups",
        return_value={"build": [], "srpm-build": []},
    )
    @patch("module_build_service.builder.GenericBuilder.create_from_module")
    @patch("module_build_service.resolver.DBResolver")
    @patch("module_build_service.resolver.GenericResolver")
    def test_set_cg_build_koji_tag_fallback_to_default(
        self, generic_resolver, resolver, create_builder, dbg
    ):
        """
        Test that build.cg_build_koji_tag fallbacks to default tag.
        """
        with app.app_context():
            base_mmd = Modulemd.ModuleStreamV2.new("base-runtime", "f27")

            scheduler_init_data()
            koji_session = mock.MagicMock()
            koji_session.newRepo.return_value = 123456

            builder = mock.MagicMock()
            builder.koji_session = koji_session
            builder.module_build_tag = {"name": "module-123-build"}
            builder.get_disttag_srpm.return_value = "some srpm disttag"
            builder.build.return_value = (
                1234,
                koji.BUILD_STATES["BUILDING"],
                "",
                "module-build-macros-1-1",
            )
            create_builder.return_value = builder

            resolver = mock.MagicMock()
            resolver.backend = "db"
            resolver.get_module_tag.return_value = "module-testmodule-master-20170109091357"
            resolver.get_module_build_dependencies.return_value = {
                "module-bootstrap-tag": [base_mmd]
            }

            with patch.object(module_build_service.resolver, "system_resolver", new=resolver):
                msg = module_build_service.messaging.MBSModule(
                    msg_id=None, module_build_id=2, module_build_state="some state")
                module_build_service.scheduler.handlers.modules.wait(
                    config=conf, session=db.session, msg=msg)
                module_build = ModuleBuild.query.filter_by(id=2).one()
                assert module_build.cg_build_koji_tag == "modular-updates-candidate"

    @pytest.mark.parametrize(
        "koji_cg_tag_build,expected_cg_koji_build_tag",
        [
            [True, "f27-modular-updates-candidate"],
            [False, None]
        ],
    )
    @patch(
        "module_build_service.builder.GenericBuilder.default_buildroot_groups",
        return_value={"build": [], "srpm-build": []},
    )
    @patch("module_build_service.builder.GenericBuilder.create_from_module")
    @patch("module_build_service.resolver.DBResolver")
    @patch("module_build_service.resolver.GenericResolver")
    @patch(
        "module_build_service.config.Config.base_module_names",
        new_callable=mock.PropertyMock,
        return_value=["base-runtime", "platform"],
    )
    def test_set_cg_build_koji_tag(
        self,
        cfg,
        generic_resolver,
        resolver,
        create_builder,
        dbg,
        koji_cg_tag_build,
        expected_cg_koji_build_tag,
    ):
        """
        Test that build.cg_build_koji_tag is set.
        """
        with app.app_context():
            base_mmd = Modulemd.ModuleStreamV2.new("base-runtime", "f27")

            scheduler_init_data()
            koji_session = mock.MagicMock()
            koji_session.newRepo.return_value = 123456

            builder = mock.MagicMock()
            builder.koji_session = koji_session
            builder.module_build_tag = {"name": "module-123-build"}
            builder.get_disttag_srpm.return_value = "some srpm disttag"
            builder.build.return_value = (
                1234,
                koji.BUILD_STATES["BUILDING"],
                "",
                "module-build-macros-1-1",
            )
            create_builder.return_value = builder

            resolver = mock.MagicMock()
            resolver.backend = "db"
            resolver.get_module_tag.return_value = "module-testmodule-master-20170109091357"
            resolver.get_module_build_dependencies.return_value = {
                "module-bootstrap-tag": [base_mmd]
            }

            with patch.object(
                module_build_service.scheduler.handlers.modules.conf,
                "koji_cg_tag_build",
                new=koji_cg_tag_build,
            ):
                with patch.object(module_build_service.resolver, "system_resolver", new=resolver):
                    msg = module_build_service.messaging.MBSModule(
                        msg_id=None, module_build_id=2, module_build_state="some state"
                    )
                    module_build_service.scheduler.handlers.modules.wait(
                        config=conf, session=db.session, msg=msg
                    )
                    module_build = ModuleBuild.query.filter_by(id=2).one()
                    assert module_build.cg_build_koji_tag == expected_cg_koji_build_tag



# -*- coding: utf-8 -*-
# Copyright (c) 2019  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>

from module_build_service import conf, log
from module_build_service.builder.KojiModuleBuilder import KojiModuleBuilder
from module_build_service.models import ModuleBuild, BUILD_STATES


def get_corresponding_module_build(session, nvr):
    """Find corresponding module build from database and return

    :param session: the SQLAlchemy database session object.
    :param str nvr: module build NVR. This is the subject_identifier included
        inside ``greenwave.decision.update`` message.
    :return: the corresponding module build object. For whatever the reason,
        if the original module build id cannot be found from the Koji build of
        ``nvr``, None will be returned.
    :rtype: :class:`ModuleBuild` or None
    """
    koji_session = KojiModuleBuilder.get_session(conf, login=False)
    build_info = koji_session.getBuild(nvr)
    if build_info is None:
        return None

    try:
        module_build_id = build_info["extra"]["typeinfo"]["module"]["module_build_service_id"]
    except KeyError:
        # If any of the keys is not present, the NVR is not the one for
        # handling Greenwave event.
        return None

    return ModuleBuild.get_by_id(session, module_build_id)


def decision_update(config, session, msg):
    """Move module build to ready or failed according to Greenwave result

    :param config: the config object returned from function :func:`init_config`,
        which is loaded from configuration file.
    :type config: :class:`Config`
    :param session: the SQLAlchemy database session object.
    :param msg: the message object representing a message received from topic
        ``greenwave.decision.update``.
    :type msg: :class:`GreenwaveDecisionUpdate`
    """
    if not config.greenwave_decision_context:
        log.debug(
            "Skip Greenwave message %s as MBS does not have GREENWAVE_DECISION_CONTEXT "
            "configured",
            msg.msg_id,
        )
        return

    if msg.decision_context != config.greenwave_decision_context:
        log.debug(
            "Skip Greenwave message %s as MBS only handles messages with the "
            'decision context "%s"',
            msg.msg_id,
            config.greenwave_decision_context,
        )
        return

    module_build_nvr = msg.subject_identifier

    if not msg.policies_satisfied:
        log.debug(
            "Skip to handle module build %s because it has not satisfied Greenwave policies.",
            module_build_nvr,
        )
        return

    build = get_corresponding_module_build(session, module_build_nvr)

    if build is None:
        log.debug(
            "No corresponding module build of subject_identifier %s is found.", module_build_nvr)
        return

    if build.state == BUILD_STATES["done"]:
        build.transition(
            conf,
            BUILD_STATES["ready"],
            state_reason="Module build {} has satisfied Greenwave policies.".format(
                module_build_nvr
            ),
        )
    else:
        log.warning(
            "Module build %s is not in done state but Greenwave tells "
            "it passes tests in decision context %s",
            module_build_nvr,
            msg.decision_context,
        )

    session.commit()



# -*- coding: utf-8 -*-
# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

""" The PollingProducer class that acts as a producer entry point for
fedmsg-hub. This class polls the database for tasks to do.
"""

import koji
import operator
from datetime import timedelta, datetime
from sqlalchemy.orm import lazyload
from moksha.hub.api.producer import PollingProducer

import module_build_service.messaging
import module_build_service.scheduler
import module_build_service.scheduler.consumer
from module_build_service import conf, models, log
from module_build_service.builder import GenericBuilder
from module_build_service.builder.KojiModuleBuilder import KojiModuleBuilder
from module_build_service.utils.greenwave import greenwave


class MBSProducer(PollingProducer):
    frequency = timedelta(seconds=conf.polling_interval)

    def poll(self):
        with models.make_session(conf) as session:
            try:
                self.log_summary(session)
                self.process_waiting_module_builds(session)
                self.process_open_component_builds(session)
                self.fail_lost_builds(session)
                self.process_paused_module_builds(conf, session)
                self.trigger_new_repo_when_stalled(conf, session)
                self.delete_old_koji_targets(conf, session)
                self.cleanup_stale_failed_builds(conf, session)
                self.sync_koji_build_tags(conf, session)
                self.poll_greenwave(conf, session)
            except Exception:
                msg = "Error in poller execution:"
                log.exception(msg)

        log.info('Poller will now sleep for "{}" seconds'.format(conf.polling_interval))

    def fail_lost_builds(self, session):
        # This function is supposed to be handling only the part which can't be
        # updated through messaging (e.g. srpm-build failures). Please keep it
        # fit `n` slim. We do want rest to be processed elsewhere
        # TODO re-use

        if conf.system == "koji":
            # We don't do this on behalf of users
            koji_session = KojiModuleBuilder.get_session(conf, login=False)
            log.info("Querying tasks for statuses:")
            res = (
                models.ComponentBuild.query.filter_by(state=koji.BUILD_STATES["BUILDING"])
                .options(lazyload("module_build"))
                .all()
            )

            log.info("Checking status for {0} tasks".format(len(res)))
            for component_build in res:
                log.debug(component_build.json())
                # Don't check tasks which haven't been triggered yet
                if not component_build.task_id:
                    continue

                # Don't check tasks for components which have been reused,
                # they may have BUILDING state temporarily before we tag them
                # to new module tag. Checking them would be waste of resources.
                if component_build.reused_component_id:
                    log.debug(
                        'Skipping check for task "{0}", '
                        'the component has been reused ("{1}").'.format(
                            component_build.task_id, component_build.reused_component_id)
                    )
                    continue

                task_id = component_build.task_id

                log.info('Checking status of task_id "{0}"'.format(task_id))
                task_info = koji_session.getTaskInfo(task_id)

                state_mapping = {
                    # Cancelled and failed builds should be marked as failed.
                    koji.TASK_STATES["CANCELED"]: koji.BUILD_STATES["FAILED"],
                    koji.TASK_STATES["FAILED"]: koji.BUILD_STATES["FAILED"],
                    # Completed tasks should be marked as complete.
                    koji.TASK_STATES["CLOSED"]: koji.BUILD_STATES["COMPLETE"],
                }

                # If it is a closed/completed task, then we can extract the NVR
                build_version, build_release = None, None  # defaults
                if task_info["state"] == koji.TASK_STATES["CLOSED"]:
                    builds = koji_session.listBuilds(taskID=task_id)
                    if not builds:
                        log.warning(
                            "Task ID %r is closed, but we found no builds in koji." % task_id)
                    elif len(builds) > 1:
                        log.warning(
                            "Task ID %r is closed, but more than one build is present!" % task_id)
                    else:
                        build_version = builds[0]["version"]
                        build_release = builds[0]["release"]

                log.info("  task {0!r} is in state {1!r}".format(task_id, task_info["state"]))
                if task_info["state"] in state_mapping:
                    # Fake a fedmsg message on our internal queue
                    msg = module_build_service.messaging.KojiBuildChange(
                        msg_id="producer::fail_lost_builds fake msg",
                        build_id=component_build.task_id,
                        task_id=component_build.task_id,
                        build_name=component_build.package,
                        build_new_state=state_mapping[task_info["state"]],
                        build_release=build_release,
                        build_version=build_version,
                    )
                    module_build_service.scheduler.consumer.work_queue_put(msg)

        elif conf.system == "mock":
            pass

    def cleanup_stale_failed_builds(self, conf, session):
        """ Does various clean up tasks on stale failed module builds
        :param conf: the MBS configuration object
        :param session: a SQLAlchemy database session
        """
        if conf.system == "koji":
            stale_date = datetime.utcnow() - timedelta(days=conf.cleanup_failed_builds_time)
            stale_module_builds = (
                session.query(models.ModuleBuild)
                .filter(
                    models.ModuleBuild.state == models.BUILD_STATES["failed"],
                    models.ModuleBuild.time_modified <= stale_date,
                )
                .all()
            )
            if stale_module_builds:
                log.info(
                    "{0} stale failed module build(s) will be cleaned up".format(
                        len(stale_module_builds))
                )
            for module in stale_module_builds:
                log.info("{0!r} is stale and is being cleaned up".format(module))
                # Find completed artifacts in the stale build
                artifacts = [
                    c for c in module.component_builds
                    if c.state == koji.BUILD_STATES["COMPLETE"]
                ]
                # If there are no completed artifacts, then there is nothing to tag
                if artifacts:
                    # Set buildroot_connect=False so it doesn't recreate the Koji target and etc.
                    builder = GenericBuilder.create_from_module(
                        session, module, conf, buildroot_connect=False
                    )
                    builder.untag_artifacts([c.nvr for c in artifacts])
                    # Mark the artifacts as untagged in the database
                    for c in artifacts:
                        c.tagged = False
                        c.tagged_in_final = False
                        session.add(c)
                state_reason = (
                    "The module was garbage collected since it has failed over {0}"
                    " day(s) ago".format(conf.cleanup_failed_builds_time)
                )
                module.transition(
                    conf,
                    models.BUILD_STATES["garbage"],
                    state_reason=state_reason,
                    failure_type="user",
                )
                session.add(module)
                session.commit()

    def log_summary(self, session):
        log.info("Current status:")
        consumer = module_build_service.scheduler.consumer.get_global_consumer()
        backlog = consumer.incoming.qsize()
        log.info("  * internal queue backlog is {0}".format(backlog))
        states = sorted(models.BUILD_STATES.items(), key=operator.itemgetter(1))
        for name, code in states:
            query = models.ModuleBuild.query.filter_by(state=code)
            count = query.count()
            if count:
                log.info("  * {0} module builds in the {1} state".format(count, name))
            if name == "build":
                for module_build in query.all():
                    log.info("    * {0!r}".format(module_build))
                    # First batch is number '1'.
                    for i in range(1, module_build.batch + 1):
                        n = len([c for c in module_build.component_builds if c.batch == i])
                        log.info("      * {0} components in batch {1}".format(n, i))

    def _nudge_module_builds_in_state(self, session, state_name, older_than_minutes):
        """
        Finds all the module builds in the `state` with `time_modified` older
        than `older_than_minutes` and adds fake MBSModule message to the
        work queue.
        """
        log.info("Looking for module builds stuck in the %s state", state_name)
        builds = models.ModuleBuild.by_state(session, state_name)
        log.info(" %r module builds in the %s state...", len(builds), state_name)
        now = datetime.utcnow()
        time_modified_threshold = timedelta(minutes=older_than_minutes)
        for build in builds:

            # Only give builds a nudge if stuck for more than ten minutes
            if (now - build.time_modified) < time_modified_threshold:
                continue

            # Pretend the build is modified, so we don't tight spin.
            build.time_modified = now
            session.commit()

            # Fake a message to kickstart the build anew in the consumer
            state = module_build_service.models.BUILD_STATES[state_name]
            msg = module_build_service.messaging.MBSModule(
                "nudge_module_builds_fake_message", build.id, state)
            log.info("  Scheduling faked event %r" % msg)
            module_build_service.scheduler.consumer.work_queue_put(msg)

    def process_waiting_module_builds(self, session):
        for state in ["init", "wait"]:
            self._nudge_module_builds_in_state(session, state, 10)

    def process_open_component_builds(self, session):
        log.warning("process_open_component_builds is not yet implemented...")

    def process_paused_module_builds(self, config, session):
        log.info("Looking for paused module builds in the build state")
        if module_build_service.utils.at_concurrent_component_threshold(config, session):
            log.debug(
                "Will not attempt to start paused module builds due to "
                "the concurrent build threshold being met"
            )
            return

        ten_minutes = timedelta(minutes=10)
        # Check for module builds that are in the build state but don't have any active component
        # builds. Exclude module builds in batch 0. This is likely a build of a module without
        # components.
        module_builds = (
            session.query(models.ModuleBuild)
            .filter(
                models.ModuleBuild.state == models.BUILD_STATES["build"],
                models.ModuleBuild.batch > 0,
            )
            .all()
        )
        for module_build in module_builds:
            now = datetime.utcnow()
            # Only give builds a nudge if stuck for more than ten minutes
            if (now - module_build.time_modified) < ten_minutes:
                continue
            # If there are no components in the build state on the module build,
            # then no possible event will start off new component builds.
            # But do not try to start new builds when we are waiting for the
            # repo-regen.
            if (
                not module_build.current_batch(koji.BUILD_STATES["BUILDING"])
                and not module_build.new_repo_task_id
            ):
                log.info("  Processing the paused module build %r", module_build)
                # Initialize the builder...
                builder = GenericBuilder.create_from_module(session, module_build, config)

                further_work = module_build_service.utils.start_next_batch_build(
                    config, module_build, session, builder)
                for event in further_work:
                    log.info("  Scheduling faked event %r" % event)
                    module_build_service.scheduler.consumer.work_queue_put(event)

            # Check if we have met the threshold.
            if module_build_service.utils.at_concurrent_component_threshold(config, session):
                break

    def trigger_new_repo_when_stalled(self, config, session):
        """
        Sometimes the Koji repo regeneration stays in "init" state without
        doing anything and our module build stucks. In case the module build
        gets stuck on that, we trigger newRepo again to rebuild it.
        """
        if config.system != "koji":
            return

        koji_session = module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder.get_session(
            config)

        for module_build in (
            session.query(models.ModuleBuild).filter_by(state=models.BUILD_STATES["build"]).all()
        ):
            if not module_build.new_repo_task_id:
                continue

            task_info = koji_session.getTaskInfo(module_build.new_repo_task_id)
            if task_info["state"] in [koji.TASK_STATES["CANCELED"], koji.TASK_STATES["FAILED"]]:
                log.info(
                    "newRepo task %s for %r failed, starting another one",
                    str(module_build.new_repo_task_id), module_build,
                )
                taginfo = koji_session.getTag(module_build.koji_tag + "-build")
                module_build.new_repo_task_id = koji_session.newRepo(taginfo["name"])
            else:
                module_build.new_repo_task_id = 0

        session.commit()

    def delete_old_koji_targets(self, config, session):
        """
        Deletes targets older than `config.koji_target_delete_time` seconds
        from Koji to cleanup after the module builds.
        """
        if config.system != "koji":
            return

        log.info("Looking for module builds which Koji target can be removed")

        now = datetime.utcnow()

        koji_session = KojiModuleBuilder.get_session(config)
        for target in koji_session.getBuildTargets():
            koji_tag = target["dest_tag_name"]
            module = session.query(models.ModuleBuild).filter_by(koji_tag=koji_tag).first()
            if (
                not module
                or module.name in conf.base_module_names
                or module.state in [
                    models.BUILD_STATES["init"],
                    models.BUILD_STATES["wait"],
                    models.BUILD_STATES["build"],
                ]
            ):
                continue

            # Double-check that the target we are going to remove is prefixed
            # by our prefix, so we won't remove f26 when there is some garbage
            # in DB or Koji.
            for allowed_prefix in config.koji_tag_prefixes:
                if target["name"].startswith(allowed_prefix + "-"):
                    break
            else:
                log.error("Module %r has Koji target with not allowed prefix.", module)
                continue

            delta = now - module.time_completed
            if delta.total_seconds() > config.koji_target_delete_time:
                log.info("Removing target of module %r", module)
                koji_session.deleteBuildTarget(target["id"])

    def cancel_stuck_module_builds(self, config, session):
        """
        Method transitions builds which are stuck in one state too long to the "failed" state.
        The states are defined with the "cleanup_stuck_builds_states" config option and the
        time is defined by the "cleanup_stuck_builds_time" config option.
        """
        log.info(
            'Looking for module builds stuck in the states "{states}" more than {days} days'
            .format(
                states=" and ".join(config.cleanup_stuck_builds_states),
                days=config.cleanup_stuck_builds_time,
            )
        )

        delta = timedelta(days=config.cleanup_stuck_builds_time)
        now = datetime.utcnow()
        threshold = now - delta
        states = [
            module_build_service.models.BUILD_STATES[state]
            for state in config.cleanup_stuck_builds_states
        ]

        module_builds = (
            session.query(models.ModuleBuild)
            .filter(
                models.ModuleBuild.state.in_(states), models.ModuleBuild.time_modified < threshold
            )
            .all()
        )

        log.info(" {0!r} module builds are stuck...".format(len(module_builds)))

        for build in module_builds:
            nsvc = ":".join([build.name, build.stream, build.version, build.context])
            log.info('Transitioning build "{nsvc}" to "Failed" state.'.format(nsvc=nsvc))

            state_reason = "The module was in {state} for more than {days} days".format(
                state=build.state, days=config.cleanup_stuck_builds_time
            )
            build.transition(
                config,
                state=models.BUILD_STATES["failed"],
                state_reason=state_reason,
                failure_type="user",
            )
            session.commit()

    def sync_koji_build_tags(self, config, session):
        """
        Method checking the "tagged" and "tagged_in_final" attributes of
        "complete" ComponentBuilds in the current batch of module builds
        in "building" state against the Koji.

        In case the Koji shows the build as tagged/tagged_in_final,
        fake "tagged" message is added to work queue.
        """
        if conf.system != "koji":
            return

        koji_session = KojiModuleBuilder.get_session(conf, login=False)

        threshold = datetime.utcnow() - timedelta(minutes=10)
        module_builds = session.query(models.ModuleBuild).filter(
            models.ModuleBuild.time_modified < threshold,
            models.ModuleBuild.state == models.BUILD_STATES["build"]
        ).all()
        for module_build in module_builds:
            complete_components = module_build.current_batch(koji.BUILD_STATES["COMPLETE"])
            for c in complete_components:
                # In case the component is tagged in the build tag and
                # also tagged in the final tag (or it is build_time_only
                # and therefore should not be tagged in final tag), skip it.
                if c.tagged and (c.tagged_in_final or c.build_time_only):
                    continue

                log.info(
                    "%r: Component %r is complete, but not tagged in the "
                    "final and/or build tags.",
                    module_build, c,
                )

                # Check in which tags the component is tagged.
                tag_dicts = koji_session.listTags(c.nvr)
                tags = [tag_dict["name"] for tag_dict in tag_dicts]

                # If it is tagged in final tag, but MBS does not think so,
                # schedule fake message.
                if not c.tagged_in_final and module_build.koji_tag in tags:
                    msg = module_build_service.messaging.KojiTagChange(
                        "sync_koji_build_tags_fake_message", module_build.koji_tag, c.package, c.nvr
                    )
                    log.info("  Scheduling faked event %r" % msg)
                    module_build_service.scheduler.consumer.work_queue_put(msg)

                # If it is tagged in the build tag, but MBS does not think so,
                # schedule fake message.
                build_tag = module_build.koji_tag + "-build"
                if not c.tagged and build_tag in tags:
                    msg = module_build_service.messaging.KojiTagChange(
                        "sync_koji_build_tags_fake_message", build_tag, c.package, c.nvr)
                    log.info("  Scheduling faked event %r" % msg)
                    module_build_service.scheduler.consumer.work_queue_put(msg)

    def poll_greenwave(self, config, session):
        """
        Polls Greenwave for all builds in done state
        :param session: SQLAlchemy DB session
        :return: None
        """
        if greenwave is None:
            return

        module_builds = (
            session.query(models.ModuleBuild)
            .filter_by(state=models.BUILD_STATES["done"]).all()
        )

        log.info("Checking Greenwave for %d builds", len(module_builds))

        for build in module_builds:
            if greenwave.check_gating(build):
                build.transition(config, state=models.BUILD_STATES["ready"])
                session.commit()



"""Add a table for virtual streams

Revision ID: 6d503efcd2b8
Revises: 79babdadc942
Create Date: 2019-04-23 19:28:52.206109

"""

# revision identifiers, used by Alembic.
revision = '6d503efcd2b8'
down_revision = '79babdadc942'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base

# Data migration imports
from module_build_service import Modulemd, conf


Base = declarative_base()

# Define the tables for the data migration
module_builds_to_virtual_streams = sa.Table(
    'module_builds_to_virtual_streams',
    Base.metadata,
    sa.Column('module_build_id', sa.Integer, sa.ForeignKey('module_builds.id')),
    sa.Column('virtual_stream_id', sa.Integer, sa.ForeignKey('virtual_streams.id')),
)


class ModuleBuild(Base):
    __tablename__ = "module_builds"
    id = sa.Column(sa.Integer, primary_key=True)
    modulemd = sa.Column(sa.String, nullable=False)
    name = sa.Column(sa.String, nullable=False)
    virtual_streams = relationship(
        'VirtualStream', secondary=module_builds_to_virtual_streams, back_populates='module_builds'
    )


class VirtualStream(Base):
    __tablename__ = 'virtual_streams'
    id = sa.Column(sa.Integer, primary_key=True)
    name = sa.Column(sa.String, nullable=False, unique=True)
    module_builds = relationship(
        'ModuleBuild', secondary=module_builds_to_virtual_streams, back_populates='virtual_streams'
    )


def upgrade():
    op.create_table(
        'virtual_streams',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name'),
    )
    op.create_table(
        'module_builds_to_virtual_streams',
        sa.Column('module_build_id', sa.Integer(), nullable=False),
        sa.Column('virtual_stream_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['module_build_id'], ['module_builds.id']),
        sa.ForeignKeyConstraint(['virtual_stream_id'], ['virtual_streams.id']),
        sa.UniqueConstraint(
            'module_build_id', 'virtual_stream_id', name='unique_module_to_virtual_stream'
        ),
    )

    bind = op.get_bind()
    session = sa.orm.Session(bind=bind)
    for build in session.query(ModuleBuild).all():
        # Only process base modules with modulemds set
        if build.name not in conf.base_module_names or not build.modulemd:
            continue

        try:
            mmd = Modulemd.ModuleStream.read_string(build.modulemd, True)
            mmd = mmd.upgrade(Modulemd.ModuleStreamVersionEnum.TWO)
        except Exception:
            # If the modulemd isn't parseable, then skip this build
            continue

        try:
            virtual_streams = mmd.get_xmd()['mbs']['virtual_streams']
        except KeyError:
            # If there are no virtual_streams configured, then just skip this build
            continue

        for virtual_stream in virtual_streams:
            virtual_stream_obj = session.query(VirtualStream).filter_by(name=virtual_stream).first()
            # Create the virtual stream entry if it doesn't exist
            if not virtual_stream_obj:
                virtual_stream_obj = VirtualStream(name=virtual_stream)
                session.add(virtual_stream_obj)
                session.commit()

            if virtual_stream_obj not in build.virtual_streams:
                build.virtual_streams.append(virtual_stream_obj)
                session.add(build)

        session.commit()

    # Always close the transaction
    session.commit()


def downgrade():
    op.drop_table('module_builds_to_virtual_streams')
    op.drop_table('virtual_streams')



# Copyright (c) 2017 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os

from mock import patch, PropertyMock

from tests import conf, clean_database
from tests.test_views.test_views import FakeSCM
import module_build_service.messaging
import module_build_service.scheduler.handlers.modules
from module_build_service import build_logs, db
from module_build_service.models import make_session, ModuleBuild, ComponentBuild
from module_build_service.utils.general import mmd_to_str, load_mmd, load_mmd_file


class TestModuleInit:
    def setup_method(self, test_method):
        self.fn = module_build_service.scheduler.handlers.modules.init
        self.staged_data_dir = os.path.join(os.path.dirname(__file__), "../", "staged_data")
        testmodule_yml_path = os.path.join(self.staged_data_dir, "testmodule_init.yaml")
        mmd = load_mmd_file(testmodule_yml_path)
        # Set the name and stream
        mmd = mmd.copy("testmodule", "1")
        scmurl = "git://pkgs.domain.local/modules/testmodule?#620ec77"
        clean_database()
        with make_session(conf) as session:
            ModuleBuild.create(
                session, conf, "testmodule", "1", 3, mmd_to_str(mmd), scmurl, "mprahl")

    def teardown_method(self, test_method):
        try:
            path = build_logs.path(1)
            os.remove(path)
        except Exception:
            pass

    @patch(
        "module_build_service.builder.KojiModuleBuilder.KojiModuleBuilder."
        "get_built_rpms_in_module_build"
    )
    @patch("module_build_service.scm.SCM")
    @patch("module_build_service.scheduler.handlers.modules.handle_stream_collision_modules")
    @patch("module_build_service.utils.submit.get_build_arches", return_value=["x86_64"])
    def test_init_basic(self, get_build_arches, rscm, mocked_scm, built_rpms):
        FakeSCM(
            mocked_scm,
            "testmodule",
            "testmodule_init.yaml",
            "620ec77321b2ea7b0d67d82992dda3e1d67055b4",
        )

        built_rpms.return_value = [
            "foo-0:2.4.48-3.el8+1308+551bfa71",
            "foo-debuginfo-0:2.4.48-3.el8+1308+551bfa71",
            "bar-0:2.5.48-3.el8+1308+551bfa71",
            "bar-debuginfo-0:2.5.48-3.el8+1308+551bfa71",
            "x-0:2.5.48-3.el8+1308+551bfa71",
            "x-debuginfo-0:2.5.48-3.el8+1308+551bfa71",
        ]

        platform_build = ModuleBuild.query.get(1)
        mmd = platform_build.mmd()
        for rpm in mmd.get_rpm_filters():
            mmd.remove_rpm_filter(rpm)
        mmd.add_rpm_filter("foo")
        mmd.add_rpm_filter("bar")
        platform_build.modulemd = mmd_to_str(mmd)
        db.session.commit()

        msg = module_build_service.messaging.MBSModule(
            msg_id=None, module_build_id=2, module_build_state="init"
        )

        with make_session(conf) as session:
            self.fn(config=conf, session=session, msg=msg)
        build = ModuleBuild.query.filter_by(id=2).one()
        # Make sure the module entered the wait state
        assert build.state == 1, build.state
        # Make sure format_mmd was run properly
        xmd_mbs = build.mmd().get_xmd()["mbs"]
        assert xmd_mbs["buildrequires"]["platform"]["filtered_rpms"] == [
            "foo-0:2.4.48-3.el8+1308+551bfa71",
            "bar-0:2.5.48-3.el8+1308+551bfa71",
        ]
        return build

    def test_init_called_twice(self):
        build = self.test_init_basic()
        old_component_builds = len(build.component_builds)
        old_mmd = load_mmd(build.modulemd)

        build.state = 4
        db.session.commit()
        build = self.test_init_basic()
        db.session.refresh(build)

        assert build.state == 1
        assert old_component_builds == len(build.component_builds)

        new_mmd = load_mmd(build.modulemd)
        assert mmd_to_str(old_mmd) == mmd_to_str(new_mmd)

    @patch("module_build_service.scm.SCM")
    @patch("module_build_service.utils.submit.get_build_arches", return_value=["x86_64"])
    def test_init_scm_not_available(self, get_build_arches, mocked_scm):
        def mocked_scm_get_latest():
            raise RuntimeError("Failed in mocked_scm_get_latest")

        FakeSCM(
            mocked_scm, "testmodule", "testmodule.yaml", "620ec77321b2ea7b0d67d82992dda3e1d67055b4")
        mocked_scm.return_value.get_latest = mocked_scm_get_latest
        msg = module_build_service.messaging.MBSModule(
            msg_id=None, module_build_id=2, module_build_state="init")
        with make_session(conf) as session:
            self.fn(config=conf, session=session, msg=msg)
        build = ModuleBuild.query.filter_by(id=2).one()
        # Make sure the module entered the failed state
        # since the git server is not available
        assert build.state == 4, build.state

    @patch(
        "module_build_service.config.Config.modules_allow_repository",
        new_callable=PropertyMock,
        return_value=True,
    )
    @patch("module_build_service.scm.SCM")
    @patch("module_build_service.utils.submit.get_build_arches", return_value=["x86_64"])
    def test_init_includedmodule(self, get_build_arches, mocked_scm, mocked_mod_allow_repo):
        FakeSCM(mocked_scm, "includedmodules", ["testmodule_init.yaml"])
        includedmodules_yml_path = os.path.join(self.staged_data_dir, "includedmodules.yaml")
        mmd = load_mmd_file(includedmodules_yml_path)
        # Set the name and stream
        mmd = mmd.copy("includedmodules", "1")
        scmurl = "git://pkgs.domain.local/modules/includedmodule?#da95886"
        with make_session(conf) as session:
            ModuleBuild.create(
                session, conf, "includemodule", "1", 3, mmd_to_str(mmd), scmurl, "mprahl")
            msg = module_build_service.messaging.MBSModule(
                msg_id=None, module_build_id=3, module_build_state="init")
            self.fn(config=conf, session=session, msg=msg)
        build = ModuleBuild.query.filter_by(id=3).one()
        assert build.state == 1
        assert build.name == "includemodule"
        batches = {}
        for comp_build in ComponentBuild.query.filter_by(module_id=3).all():
            batches[comp_build.package] = comp_build.batch
        assert batches["perl-List-Compare"] == 2
        assert batches["perl-Tangerine"] == 2
        assert batches["foo"] == 2
        assert batches["tangerine"] == 3
        assert batches["file"] == 4
        # Test that the RPMs are properly merged in xmd
        xmd_rpms = {
            "perl-List-Compare": {"ref": "4f26aeafdb"},
            "perl-Tangerine": {"ref": "4f26aeafdb"},
            "tangerine": {"ref": "4f26aeafdb"},
            "foo": {"ref": "93dea37599"},
            "file": {"ref": "a2740663f8"},
        }
        assert build.mmd().get_xmd()["mbs"]["rpms"] == xmd_rpms

    @patch("module_build_service.models.ModuleBuild.from_module_event")
    @patch("module_build_service.scm.SCM")
    @patch("module_build_service.utils.submit.get_build_arches", return_value=["x86_64"])
    def test_init_when_get_latest_raises(
            self, get_build_arches, mocked_scm, mocked_from_module_event):
        FakeSCM(
            mocked_scm,
            "testmodule",
            "testmodule.yaml",
            "7035bd33614972ac66559ac1fdd019ff6027ad22",
            get_latest_raise=True,
        )
        msg = module_build_service.messaging.MBSModule(
            msg_id=None, module_build_id=2, module_build_state="init")
        with make_session(conf) as session:
            build = session.query(ModuleBuild).filter_by(id=2).one()
            mocked_from_module_event.return_value = build
            self.fn(config=conf, session=session, msg=msg)
            # Query the database again to make sure the build object is updated
            session.refresh(build)
            # Make sure the module entered the failed state
            assert build.state == 4, build.state
            assert "Failed to get the latest commit for" in build.state_reason
